-- FELIX -- an approximate emulator for Felix-C-256 assembly (ASSIRIS) jobs
-- Copyright (c) 2023--2024 Alexandru Dan Corlan, MD, PhD (alexandru@corlan.net)
--
-- This is release V0.3.0, January 21, 2024
--
-- Part of the RASSIRIS project, to simulate Felix/Assiris-like computers
--
-- http://dan.corlan.net/software/emulators/felix-assiris/
--
-- Detailed documentation is at:
--
-- http://dan.corlan.net/software/emulators/felix-assiris/rassiris.pdf
--
-- This code is released under the GNU General Public Licence version 2
-- 
-- INSTALLATION. 
-- On LINUX. Install gnat, the GNU Ada Translator. On debian: apt install gnat
-- Copy this file in a terminal and run: gnatchop -w -r felix.ada ; gnatmake felix
-- The resulting binary, felix, can be run locally or copied to /usr/local/bin
-- for general availability on your system.
--
-- On other systems, there are two variants:
--   (1) A. Get linux. B. Do the above
--   (2) Experiment on your own with gnat installation and felix compilation
--       then add a paragraph to the above on how to do it, and release
--
-- RUNNING.

-- The simulator is a linux executable, felix, that takes cards on
-- stdin and prints on stdout. Try: ./felix <hello_world.assiris
-- for an example.
-- 

-- HISTORY.
-- Version 0.1.0  November 11, 2023. https://github.com/dancorl/felix-assiris/archive/refs/tags/v0.1.0.tar.gz
-- Version 0.1.1  November 13, 2023. https://github.com/dancorl/felix-assiris/archive/refs/tags/v0.1.1.tar.gz
--                                   https://zenodo.org/records/10117525
-- Version 0.2.0  December  1, 2023  http://dan.corlan.net/software/emulators/felix-assiris/felix-assiris-0.2.0.tar.gz
-- Version 0.3.0  January  21, 2024  (this version)

with Text_Io; use Text_Io;
with Ada.Integer_Text_Io; use Ada.Integer_Text_Io;
with Ada.Command_Line; use Ada.Command_Line;
with Ada.Calendar; use Ada.Calendar;
with Ada.Numerics.Float_Random; use Ada.Numerics.Float_Random;

procedure Felix is
   
   package Real_IO is new Float_IO(Long_Float); use Real_IO;
   
   function Heading(N: Natural; Title: String; Pre: Natural:= 0; Post: Natural := 0) return String;
   
   -- CONFIGURATION VARIABLES
   
   Verbosity: Integer:= 10;
   Show_Opcodes: Boolean:= False;
   Show_Job_Banner: Boolean:= True;
   Show_All_Assertions: Boolean:= True;
   Show_Failed_Assertions: Boolean:= True;
   Show_Run_Output_Limits: Boolean:= True;
   Show_Exit_Reason: Boolean:= True;
   Show_Run_Statistics: Boolean:= True;
   Show_Assembly: Boolean:= True;
   Show_Directives: Boolean:= True;
   
   type Timing_Options is (Original, Native);
   
   Timing_Conf: Timing_Options:= Native;
   
   Max_Allowed_Asm_Error_Level: Natural:= 10;
   Max_Allowed_Exec_Error_Level: Natural:= 10;
   Max_Allowed_Ctrl_Error_Level: Natural:= 10;
   
   -- Error Counters and max severity
   
   Nr_Asm_Errors: Natural:= 0;
   Max_Asm_Error_Level: Natural:= 0;
   
   Nr_Exec_Errors: Natural:= 0;
   Max_Exec_Error_Level: Natural:= 0;
   
   Nr_Ctrl_Errors: Natural:= 0;
   Max_Ctrl_Error_Level: Natural:= 0;
   
   N_Asserts: Natural:= 0;
   N_Failed_Asserts: Natural:= 0;
   
   -- BASIC TYPES
   
   type Bit is mod 2;
   type Dbit is mod 4;
   type Tbit is mod 8;
   type Nibble is mod 16;
   type Bit7 is mod 2**7;
   type Byte is mod 2**8;
   type Half is mod 2**16;
   type Word is mod 2**32;
   type Doub is mod 2**64;

   type Opcod is mod 2**7;
   type Iflag is mod 2;
   type Xflag is mod 2;
   type Bfld is mod 2**3;
   type Qfld is mod 2**4;
   
   type Instr is record
      I: Iflag;
      B: Bfld;
      Q: Qfld;
      Op: Opcod;
      X: Xflag;
      D: Half;
   end record;
   pragma Pack(Instr);
   for Instr'Size use 32;
   
   type Memory_Bytes is array(Half range <>) of Byte;
   
   type Psw_Type is record
      Unused1: Half;
      Ai: Half; -- this can be `half' only in CC16. In CC24 it must be at least 18 if not 20 bits
      -- the last two of which are always 00
      Ds, Di, Dd, Db: Bit; -- masti de depasire
      Z,S,D,C: Bit; -- indicatori de conditie
      Iis,Iic,Ice: Bit; -- masti de intrerupere
      Np: Bit; -- indicator de mod
      Ca: Nibble; 
      Unused2: Byte;
      Pm: Bit;
      Nit: Bit7;
   end record;
   pragma Pack(Psw_Type);
   for Psw_Type'Size use 64;
   
   -- THE MEMORY
   
   M: Memory_Bytes(Half);
   for M'Alignment use 8;
   
   -- THE REGISTERS; there could be more register banks, currently only one
   --                the registers are addressed like memory and corespond
   --                to the first 64 bytes of memory. 
   
   type Bytes_Register is new Memory_Bytes(0..3);
--   for Bytes_Register'Alignment use 8;
   type    Halfs_Register is array(0..1) of Half;
--   for Halfs_Register'Alignment use 8;
   type Words_Register is new Word;
--   for Words_Register'Alignment use 8;
   type Shorts_Register is array(0..1) of Short_Integer; -- assumed 16bit
   
   Max_Register: constant Half := 16;
   
   subtype Register is Half range 0..Max_Register-1;
   subtype Dregister is Half range 0..(Max_Register/2)-1;
   
   type Register_Bank is array(Register) of Words_Register;
   
   Rw: Register_Bank;
   for Rw'Address use M'Address;
   
   Rh: array(Register) of Halfs_Register;
   for Rh'Address use M'Address;
   
   Rb: array(Register) of Bytes_Register;
   for Rb'Address use M'Address;
   
   Ri: array(Register) of Integer;
   for Ri'Address use M'Address;
   
   Rs: array(Register) of Shorts_Register;
   for Rs'Address use M'Address;
   
   Rf: array(Register) of Float;
   for Rf'Address use M'Address;
   
   Rg: array(Dregister) of Long_Float; --
   for Rg'Address use M'Address;
   
   Rl: array(Dregister) of Long_Integer; --
   for Rl'Address use M'Address;
   
   Ru: array(Dregister) of Doub; --
   for Ru'Address use M'Address;
   
   Halted: Boolean:= False;
   In_Error: Boolean:= False;
   Psw: Psw_Type;
   Pc: Half  renames Psw.Ai;
   
   Pswbytes: Memory_Bytes(0..7);
   for Pswbytes'Address use Psw'Address;
      
   -- INS -- a register where we read the nextr instruction to execute or assemble
   
   Inw: Word;
   Inb: Memory_Bytes(0..3);
   for Inb'Address use Inw'Address;
   Ins: Instr;
   for Ins'Address use Inw'Address;
   
   Asm: Half:= 64; -- after the registers..
   Card_Addr: Half:= Asm; -- the asm before this card was assembled
   Run_Entry: Half:= 64; -- where the program will run from, given by the END directive
   
   -- JOB CONTROL CONDITIONS
   
   Injob: Boolean:= False;
   Incompile: Boolean:= False;
   
   -- some utilities
   
      
   Ai: constant Integer:= Character'Pos('a');
   Amai: constant Integer:= Character'Pos('A');
   Zeroi: constant Integer:= Character'Pos('0');
   
   
   
   
   function To_Hex_Digit(X: Natural) return Character is
      
   begin
      if X<10 then
	 return Character'Val(X + Zeroi);
      else
	 return Character'Val(X - 10 + Amai);
      end if;
   end To_Hex_Digit;
   
   function To_Hex_Half(X: Half) return String is
	 
      Rv: String(1..4);
	 
   begin
      Rv(1) := To_Hex_Digit(Natural(X / 2**12));
      Rv(2) := To_Hex_Digit(Natural((X / 2**8) mod 2**4));
      Rv(3) := To_Hex_Digit(Natural((X / 2**4) mod 2**4));
      Rv(4) := To_Hex_Digit(Natural(X mod 2**4));
      return Rv;
   end To_Hex_Half;
   
   function To_Hex_byte(X: Byte) return String is
      
      Rv: String(1..2);
      
   begin
      Rv(1) := To_Hex_Digit(Natural(X / 2**4));
      Rv(2) := To_Hex_Digit(Natural(X mod 2**4));
      return Rv;
   end To_Hex_Byte;
      
   function To_Hex_Word(X: Word) return String is
      Rv: String(1..8);
      Y: array(1..4) of Byte;
      for Y'Address use X'Address;
      
   begin 
      Rv(1..2) := To_Hex_Byte(Y(1));
      Rv(3..4) := To_Hex_Byte(Y(2));
      Rv(5..6) := To_Hex_Byte(Y(3));
      Rv(7..8) := To_Hex_Byte(Y(4));
      
      return Rv;
   end To_Hex_Word;
      
   function To_Upper(X: String) return String is
      
      Rv: String(1..X'Length);
      
   begin
      for I in X'Range loop
	 if X(I) <= 'z' and X(I) >= 'a' then
	    Rv(I-X'First+1) := Character'Val(Character'Pos(X(I)) + Amai - Ai);
	 else
	    Rv(I-X'First+1) := X(I);
	 end if;
      end loop;
      return Rv;
   end To_Upper;
      
   function Is_White(C: Character) return Boolean is
      
   begin
      return C=' ' or C=Ascii.Ht;
   end Is_White;
   
   function Is_Alpha(C: Character) return Boolean is
      
   begin
      return (C >= 'a' and C <= 'z') or (C >='A' and C <= 'Z');
   end Is_Alpha;
   
   function Is_Decimal(C: Character) return Boolean is
      
   begin
      return C >= '0' and then C<='9';
   end Is_Decimal;
   
   function Is_Alphanum(C: Character) return Boolean is
      
   begin
      return Is_Alpha(C) or Is_Decimal(C);
   end Is_Alphanum;
   
   -- The SYMBOL TABLE
   
   Labelmax: constant Natural := 8;
   Symbol_Table_Max: constant Natural := 10_000;
   Symbol_Not_Found: constant Natural := Symbol_Table_Max + 1;
   subtype Label_String is String(1..Labelmax);
   
   type Symbol_Table_Entry is record
      Label: Label_String;
      Value: Word;
   end record;
   
   subtype St_Cursor is Natural range 1..Symbol_Table_Max;

   St: array(St_Cursor) of Symbol_Table_Entry;
   St_Next: St_Cursor:= 1;
   
   -- The LINK TABLE
   
   subtype Addr is Half;
   
   Link_Table_Max: constant Natural := 20_000;
   
   type Link_Table_Entry is record
      Label: Label_String;
      Location: Addr;
   end record;
   
   subtype Lt_Cursor is Natural range 1..Link_Table_Max;
   Lt: array(Lt_Cursor) of Link_Table_Entry;
   Lt_Next: Lt_Cursor:= 1;
   
   procedure Add_To_Lt(Lb: String; La: Addr) is
      
      Ls: Label_String := ( others => ' ');
      
   begin
      Ls(Ls'First..Ls'First+Lb'Length-1) := Lb;
      Lt(Lt_Next).Label := Ls;
      Lt(Lt_Next).Location := La;
      Lt_Next := Lt_Next+1;
   end Add_To_Lt;
   
   procedure Print_Symbol_Table is
      
   begin
      Put(Heading(1, "SYMBOL TABLE", 1, 1));
      for K in 1..St_Next-1 loop
	 Put_Line(St(K).Label & " " & To_Hex_Half(Half(St(K).Value)) & " " & Word'Image(St(K).Value));
      end loop;
   end Print_Symbol_Table;
   
   function String_To_Label(Ll: String) return Label_String is
      
      Ls: Label_String := (others => ' ');
      L: String:= To_Upper(Ll);

   begin 
      if L'Length > Labelmax then
	 Ls(1..Labelmax) := L(L'First..L'First+Labelmax-1);
      else
	 Ls(1..L'Length) := "" & L;
      end if;
      return Ls;
   end String_To_Label;
   
   function Find_In_Table(L: string) return Natural is
      -- return cursor in table where the label is or
      -- symbol_table_max+1 if not found
      Ls: Label_String:= String_To_Label(L);
   begin
      for K in 1..St_Next-1 loop
	 if Ls=St(K).Label then
	    return K;
	 end if;
      end loop;
      return Symbol_Not_Found;
   end Find_In_Table;
   
   procedure Add_To_Table(L: String; V: Word) is
      
      Ls: Label_String := String_To_Label(L);
      Ni: Natural:= Find_In_Table(L);
      
   begin
      if Ni=Symbol_Not_Found then
	 St(St_Next).Label := Ls;
	 St(St_Next).Value := V;
	 St_Next := St_Next+1;
      else
	 St(Ni).Value := V; -- we just replace the value, but this should be reported
      end if;
   end Add_To_Table;
   
   -- THE SYSTEM MONITOR
   
   -- THE ASSERTIONS TABLE (and not only assertions perhaps) this
   --  structure is passed by the assembler to the running monitor
   --  (some) host instructions, such as ASEQ, work by extracting data
   --  from this structure and doing their work, when specific
   --  instructions are encountered. It probably corresponds, broadly,
   --  to debug information, or debug data structures that should be present
   --  in the monitor (or whatever it will be called), passed via the object file
   
   type Assertion_Type is (Word_Eq, Float_Eq, Double_Eq, Int_Word_Eq, C, NC, Z, NZ, S, NS, D, ND, CT, CF);
   
   type Asrtab_Ent is record
      Adr: Half; -- this is the address of the PC where the assertion is encountered
      Reg: Qfld; -- the register to test
      Assertt: Assertion_Type; -- what kind of test
      Ival: Integer;
      Wval: Word;
      Hval: Half;
      Fval: Float;
      Dval: Long_Float;
   end record;
   
   Max_Assertion: constant Integer:= 1_000;
   
   subtype Assertion is Integer range 1..Max_Assertion;
   
   Asr: array(Assertion) of Asrtab_Ent;
   
   Next_Asr: Assertion:= 1;
   
   -- job data
   
   Rungen: Generator;
   
   subtype Id is String(1..8);
   Jn, Pn, An, Runn: Id:= (others => ' ');
   Jstart, Jrunstart, Jrunend, Jend: Time;
   
   -- THE CARD READER (reads from standard_input).
   
   subtype Card is String(1..80); -- a line
   
   Incard: Card;
   Inlen: Natural;
   No_More_Cards: Boolean:= False;
   
   Cardno: Natural:= 0;
   
   procedure Set_Verbosity(V: Integer) is
      
   begin
      Verbosity := V;
      if Verbosity = 0 then
	 Show_Opcodes:= False;
	 Show_Job_Banner:= False;
	 Show_All_Assertions:= False;
	 Show_Failed_Assertions:= False;
	 Show_Run_Output_Limits:= False;
	 Show_Exit_Reason:= False;
	 Show_Run_Statistics:= False;
	 Show_Assembly:= False;
	 Show_Directives:= False;
      end if;
      if Verbosity >= 10 then
	 Show_Opcodes:= False;
	 Show_Job_Banner:= False;
	 Show_All_Assertions:= True;
	 Show_Failed_Assertions:= True;
	 Show_Run_Output_Limits:= True;
	 Show_Exit_Reason:= True;
	 Show_Run_Statistics:= True;
	 Show_Assembly:= True;
	 Show_Directives:= True;
      end if;
   end Set_Verbosity;
   
   procedure ASM_Error(Msg: String; Level: Natural := 10) is
      
   begin
      Put_Line(Standard_Error, 
	       "ASM ERR: " & Msg & " in card nr: " & 
		 Integer'Image(Cardno) & " : " & Ascii.Lf & 
		 "         " & Incard(1..Inlen) & Ascii.Lf);
      if Level > Max_Asm_Error_Level then
	 Max_Asm_Error_Level := Level;
      end if;
      Nr_Asm_Errors := Nr_Asm_Errors + 1;
      if Max_Asm_Error_Level > Max_Allowed_Asm_Error_Level then
	 In_Error := True; -- this should be `unrecoverable error' when error level is lower than ...
      end if;
   end ASM_Error;
   
   procedure Ctrl_Error(Msg: String; Level: Natural := 10) is
      
   begin
      Put_Line(Standard_Error, 
	       "CTRL ERR: " & Msg & " in card nr: " & 
		 Integer'Image(Cardno) & " : " & Ascii.Lf & 
		 "         " & Incard(1..Inlen) & Ascii.Lf);
      if Level > Max_Ctrl_Error_Level then
	 Max_Ctrl_Error_Level := Level;
      end if;
      Nr_Ctrl_Errors := Nr_Ctrl_Errors + 1;
      if Max_Ctrl_Error_Level > Max_Allowed_Ctrl_Error_Level then
	 In_Error := True; -- this should be `unrecoverable error' when error level is lower than ...
      end if;
   end Ctrl_Error;
   
   
   procedure Read_Next_Card(Fin: File_Type) is
      
   begin
      Incard := ( others => ' '); -- this is card semantics..
      Cardno := Cardno + 1;
      declare
	 Line: String:= Get_Line(Fin);
      begin
	 if Line'Length>80 then -- ignore beyond 80 characters
	    Incard := Line(Line'First..Line'First+79);
	    Inlen := 80;
	 else
	    Incard(Incard'First..Incard'First+Line'Length-1) := Line;
	    Inlen := Line'Length;
	 end if;
      end;
   exception
      when End_Error => No_More_Cards := True;
   end Read_Next_Card;
   
   -- THE CARD PARSER AND PRINTER
   
   type Span is array(0..1) of Natural;

   Card_Type: Character; -- ' ' unlabeled instruction, '.' control card, '*' comment, 'A' labeled instruction, '$' error
   Label, Mnemo, R, R2, Arg: Span := (others => 0);
   
   Maxarg: constant Natural:= 1000; -- we have enough memory
   
   type Arg_Type is ( Fx, Dx, Int, Str, Ch, Reg, Longstr, Lbl, Unknown );
   
   type Arg_Component is record
      T: Arg_Type;
      Val:  Span;
      Name: Span;
      Len:  Natural;
      Sval: Card; -- another card
      Hval: Half;
      Ival: Integer;
      Fval: Float;
      Dval: Long_Float;
   end record;
   
   subtype Argn is Natural range 1..Maxarg;
   
   type Args_Type is array(Argn) of Arg_Component;
   
   Args: Args_Type;
   Nextarg: Argn:= 1;
   
   
   -- the following will treat all expressions
   -- that are expected to be integers
   
   
   function Is_Declit(X: String) return Boolean is
      
   begin
      for K in X'Range loop
	 if K=X'First and then X(K)='-' then
	    null;
	 else
	    if X(K)<'0' or else X(K)>'9' then
	       return False;
	    end if;
	 end if;
      end loop;
      return True;
   end Is_Declit;
   
   function Is_Starred(X: String) return Boolean is
      
   begin
      for K in X'Range loop
	 if K=X'First then 
	    if X(K)/='*' then
	       return False;
	    end if;
	 else
	    if X(K)<'0' or else X(K)>'9' then
	       return False;
	    end if;
	 end if;
      end loop;
      return True;
   end Is_Starred;
   
   
   function Is_Charlit(X:String) return Boolean is
      
   begin
      return X(X'First)='C' and then X(X'First+1)='''
	and then X(X'Last)=''';
   end Is_Charlit;
   
   function Is_Hexlit(X:String) return Boolean is
      
   begin
      if X(X'First)='X' and then X(X'First+1)='''
	and then X(X'Last)=''' then
	 for K in X'First+2..X'Last-1 loop
	    if not ((X(K)>='A' and X(K)<='F') or
		      (X(K)>='0' and X(K)<='9')) then
	       return False;
	    end if;
	 end loop;
	 return True;
      else
	 return False;
      end if;
   end Is_Hexlit;
   
   function Is_Fslit(X:String) return Boolean is
      
   begin
      if X(X'First)='F' and then X(X'First+1)='S' and then X(X'First+2)='''
	and then X(X'Last)=''' then
	 return True;
      else
	 return False;
      end if;
   end Is_Fslit;

   function Is_Fllit(X:String) return Boolean is
      
   begin
      if X(X'First)='F' and then X(X'First+1)='L' and then X(X'First+2)='''
	and then X(X'Last)=''' then
	 return True;
      else
	 return False;
      end if;
   end Is_Fllit;

   
   function Is_Label(X: String) return Boolean is
      
   begin
      for K in X'Range loop
	 if K=X'First and then not Is_Alpha(X(K)) then
	    return False;
	 end if;
	 if not Is_Alphanum(X(K)) then
	    return False;
	 end if;
      end loop;
      return True;
      --      return Find_In_Table(X) /= Symbol_Not_Found;
   end Is_Label;
   
   function To_Half(X: String) return Half is
      
      Rv: Half:= 0;
      
   begin
      if Is_Declit(X) then
	 return Half(Integer'Value(X) mod 2**16);
      elsif Is_Label(X) then
	 --	 Put_Line("symbol is: |" & X & "|");
	 --	 Put_Line("table location is: " & St_Cursor'Image(Find_In_Table(X)));
	 return Half(St(Find_In_Table(X)).Value);
      elsif Is_Charlit(X) then
	 for K in X'First+2..X'Last-1 loop
	    Rv := Rv * 256 + Character'Pos(X(K));
	 end loop;
	 return Rv;
      elsif Is_Hexlit(X) then
	 for K in X'First+2..X'Last-1 loop
	    if X(K)>='0' and X(K)<='9' then
	       Rv := Rv * 16 + Character'Pos(X(K)) - Half(Zeroi);
	    else
	       Rv := Rv * 16 + Character'Pos(X(K)) - Half(Amai) + 10;
	    end if;
	 end loop;
	 return Rv;	    
      else
	 null;
	 --	 Put_Line(Standard_Error, "I do not understand: " & X);
      end if;
      return 0;
   end To_Half;
   
   
   function Memread2(A: Half) return Short_Integer is
      
      Lptr: array(Half range 0..2**15-1) of Short_Integer;
      for Lptr'Address use M'Address;
      
   begin
      -- check that a alignment is 8 and give an error
      -- this is an automatic alignment
      return Lptr(A/2);
   end Memread2;
   
   function Memread8(A: Half) return Long_Integer is
      
      Lptr: array(Half range 0..2**13-1) of Long_Integer;
      for Lptr'Address use M'Address;
      
   begin
      -- check that a alignment is 8 and give an error
      -- this is an automatic alignment
      return Lptr(A/8);
   end Memread8;
   
   function Memread8(A: Half) return Long_Float is
      
      Lptr: array(Half range 0..2**13-1) of Long_Float;
      for Lptr'Address use M'Address;
      
   begin
      -- check that a alignment is 8 and give an error
      -- this is an automatic alignment
      return Lptr(A/8);
   end Memread8;
   
   function Memread4(A: Half) return Integer is
      
      Lptr: array(Half range 0..2**14-1) of Integer;
      for Lptr'Address use M'Address;
      
   begin
      -- check that a alignment is 4 and give an error
      -- this is an automatic alignment
      return Lptr(A/4);
   end Memread4;
   
   procedure Memstor4(A: Half; V: Integer) is
      
      Lptr: array(Half range 0..2**14-1) of Integer;
      for Lptr'Address use M'Address;
      
   begin
      -- check that a alignment is 4 and give an error
      -- this is an automatic alignment
      Lptr(A/4) := V;
      return ;
   end Memstor4;
   
   procedure Memstor8(A: Half; V: Long_Integer) is
      
      Lptr: array(Half range 0..2**13-1) of Long_Integer;
      for Lptr'Address use M'Address;
      
   begin
      -- check that a alignment is 4 and give an error
      -- this is an automatic alignment
      Lptr(A/8) := V;
      return ;
   end Memstor8;
   
   procedure Memstor4(A: Half; V: Float) is
      
      Lptr: array(Half range 0..2**14-1) of Float;
      for Lptr'Address use M'Address;
      
   begin
      -- check that a alignment is 4 and give an error
      -- this is an automatic alignment
      Lptr(A/4) := V;
      return ;
   end Memstor4;
   
   procedure Memstor8(A: Half; V: Long_Float) is
      
      Lptr: array(Half range 0..2**13-1) of Long_Float;
      for Lptr'Address use M'Address;
      
   begin
      -- check that a alignment is 8 and give an error
      -- this is an automatic alignment
      Lptr(A/8) := V;
      return ;
   end Memstor8;

   
   procedure Memstor2(A: Half; V: Integer) is
      
      Lptr: array(Half range 0..2**15-1) of Short_Integer;
      for Lptr'Address use M'Address;
      
   begin
      -- check that a alignment is 2 and give an error
      -- this is an automatic alignment
      Lptr(A/2) := Short_Integer(V);
      return ;
   end Memstor2;
   
   function Memread4(A: Half) return Float is
      
      Lptr: array(Half range 0..2**14-1) of Float;
      for Lptr'Address use M'Address;
      
   begin
      -- check that a alignment is 4 and give an error
      -- this is an automatic alignment
      return Lptr(A/4);
   end Memread4;

   
   procedure Parse_Arg is
      
      procedure Clear_Arg(A: Argn) is
	 
      begin
	 Args(A).Val := ( others => 0);
	 Args(A).T := Unknown;
	 Args(A).Name := (others => 0);
	 Args(A).Ival := 0;
	 Args(A).Fval := 0.0;
	 Args(A).Dval := 0.0;
      end Clear_Arg;
      
      N: Natural;
      
   begin
      Nextarg:= 1;
      if Arg(0)=0 then
	 return ;
      end if;
      N := Arg(0);
      
      loop
	 Clear_Arg(Nextarg);
	 Args(Nextarg).Val(0) := N;
	 while N<=Arg(1) and then Incard(N)/=',' loop
	    if Incard(N)=''' then 
	       N := N+1;
	       while N<= Arg(1) and then Incard(N)/=''' loop
		  N := N+1;
	       end loop;
	       if N>Arg(1) then -- we have a tick problem
		  Asm_Error("Syntax error");
	       end if;
	    end if;
	    N := N+1;
	 end loop;
	 Args(Nextarg).Val(1) := N-1;
	 Nextarg := Nextarg+1;
	 if N>Arg(1) then
	    exit;
	 end if;
	 N := N+1; -- it is ','
      end loop;
      -- parse each arg in turn
      for A in 1..Nextarg-1 loop
	 -- does it have a name?
	 for C in Args(A).Val(0)..Args(A).Val(1) loop
	    if not Is_Alphanum(Incard(C)) then
	       if Incard(C)=':' then -- it has an alphanum name
		  Args(A).Name(0) := Args(A).Val(0);
		  Args(A).Name(1) := C-1;
		  Args(A).Val(0) := C+1;
	       end if;
	       exit;
	    end if;
	 end loop;
	 -- now, parse the value
	 declare
	    X: String:= "" & Incard(Args(A).Val(0)..Args(A).Val(1));
	    Chi: Integer;
	 begin
	    if Is_Declit(X) then
	       Args(A).Ival := Integer'Value(X);
	       Args(A).Hval := Half(Args(A).Ival mod 2**16);
	       Args(A).T := Int;
	    elsif Is_Fslit(X) then
	       Args(A).Fval := Float'Value(X(X'First+3..X'Last-1));
	       Args(A).T := Fx;
	    elsif Is_Fllit(X) then
	       Args(A).Dval := Long_Float'Value(X(X'First+3..X'Last-1));
	       Args(A).T := Dx;
	    elsif Is_Starred(X) then
	       Args(A).T := Reg;
	       Args(A).Ival := Integer'Value(Incard(Args(A).Val(0)+1 .. Args(A).Val(1)));
	       Args(A).Hval := Half(Args(A).Ival mod 2**16);
	    elsif Is_Label(X) then
	       if Find_In_Table(X) = Symbol_Not_Found then
		  Args(A).Hval := 0;
	       else
		  Args(A).Hval := Half(St(Find_In_Table(X)).Value);
	       end if;
	       Args(A).Ival := Integer(Args(A).Hval);
	       Args(A).T := Lbl;
	    elsif X(X'First)=''' and X(X'Last)=''' then
	       Args(A).T := Str;
	       Args(A).Sval(1..X'Length-2) := "" & X(X'First+1..X'Last-1);
	       Args(A).Len := X'Length-2;
	    elsif Is_Charlit(X) then
	       Chi := 0;
	       for K in X'First+2..X'Last-1 loop
		  Chi := Chi * 256 + Character'Pos(X(K));
	       end loop;
	       Args(A).Ival := Chi;
	       Args(A).Hval := Half(Chi mod 2**16);
	       Args(A).T := Ch;
	    elsif Is_Hexlit(X) then
	       Chi := 0;
	       for K in X'First+2..X'Last-1 loop
		  if X(K)>='0' and X(K)<='9' then
		     Chi := Chi * 16 + Character'Pos(X(K)) - Zeroi;
		  else
		     Chi := Chi * 16 + Character'Pos(X(K)) - Amai + 10;
		  end if;
	       end loop;
	       Args(A).Ival := Chi;
	       Args(A).Hval := Half(Chi mod 2**16);
	       Args(A).T := Int;
	    else
	       --	       Put_Line(Standard_Error, "I do not understand: " & X);
	       null;
	    end if;
	 end;
	 
      end loop;
   end Parse_Arg;
   

   procedure Parse_Card is
      
      Instring: Boolean:= False;
      I,J: Natural;

   begin
      Card_Type := '*';
      Label := (others => 0);
      Mnemo := (others => 0);
      R := (others => 0);
      R2 := (others => 0);
      Arg := (others => 0);
      
      Nextarg := 1;
      
      if Inlen < 2 then
	 return;
      end if;     
      
      case Incard(1) is
	 when '*' =>
	    Card_Type := '*';
	    return;
	 when '.' =>
	    Card_Type := '.';
	 when ' ' | Ascii.Ht =>
	    for K in 2..Inlen loop
	       if Is_Alpha(Incard(K)) then
		  Card_Type := ' ';
		  exit ;
	       end if;
	    end loop;
	    -- a space containing card, thus a comment; ignore
	 when 'A'..'Z'|'a'..'z' =>
	    Card_Type := 'A'; -- labeled 
	 when others =>
	    null; -- ignore, treat as comment
      end case;
      
      if Card_Type = '.' then
	 I := 2;
	 while I<70 and then Is_White(Incard(I)) loop
	    I := I+1;
	 end loop;
	 if I>=70 then -- error in fact
	    Asm_Error("Unrecognised control card.");
	    Card_Type := '*';
	    return; -- ignore, perhaps not best idea
	 end if;
	 J := I;
	 while J<I+10 and then Is_Alpha(Incard(J)) loop
	    J := J+1;
	 end loop;
	 if J>=I+10 then
	    Asm_Error("Unrecognised control card.");
	    Card_Type := '*';
	    return; -- ignore, perhaps not best idea
	 end if;
	 Mnemo(0) := I;
	 Mnemo(1) := J-1;
	 while J<72 and then Is_White(Incard(J)) loop
	    J := J+1; -- could be white in rest, that's ok
	 end loop;
	 if J<72 then
	    I := J;
	    Instring := False;
	    while J<72 and then not (Is_White(Incard(J)) and then not Instring) loop
	       if Incard(J)=''' then
		  Instring := not Instring;
	       end if;
	       J := J+1;
	    end loop;
	    if Is_White(Incard(J)) then
	       Arg(0) := I;
	       Arg(1) := J-1;
	    else
	       Asm_Error("Unrecognised control card.");
	       Card_Type := '*';
	       return; -- ignore, perhaps not best idea
	    end if;
	 end if;
	 Parse_Arg;
	 return;
      end if;
      
      if Card_Type = 'A' or else Card_Type = ' ' then
	 
 	 if Card_Type = 'A' then
	    Label(0) := 1;
	    for K in 1..9 loop -- maximum label size
	       if Is_White(Incard(K)) then
		  Label(1) := K-1;
		  exit;
	       end if;
	    end loop;
	    
	    if Label(1) = 0 then
	       -- something is wrong, raise an error
	       Asm_Error("Label too long");
	       -- ignored
	       Card_Type := '$';
	       return;
	    end if;
	 end if;
	 
	 for K in Label(1)+1 .. 16 loop -- the opcode, directive etc must start within first 16 characters
	    if Is_Alpha(Incard(K)) then -- mnemo must be here
	       Mnemo(0) := K;
	       for M in K..K+10 loop
		  if not Is_Alphanum(Incard(M)) then
		     Mnemo(1) := M-1;
		     exit;
		  end if;
	       end loop;
	       if Mnemo(1) = 0 then	
		  -- something is wrong, raise an error
		  Asm_Error("Mnemonic too long");
		  -- ignored
		  Card_Type := '$';
		  return;
	       end if;
	       if Incard(Mnemo(1)+1)=',' then
		  R(0) := Mnemo(1)+2;
		  J := Mnemo(1) + 3;
		  loop
		     if Is_White(Incard(J)) or J=Inlen or Incard(J)=',' then
			R(1) := J-1;
			exit;
		     end if;
		     J := J+1;
		     if J-R(0)>3 then
			Asm_Error("Reg too long");
			-- ignored
			Card_Type := '$';
			return;
		     end if;
		  end loop; 
	       end if;
	       
	       if Incard(R(1)+1)=',' then
		  R2(0) := R(1)+2;
		  J := R(1)+3;
		  loop
		     if Is_White(Incard(J)) or J=Inlen then
			R2(1) := J-1;
			exit;
		     end if;
		     J := J+1;
		     if J-R(0)>3 then
			Asm_Error("Second reg too long");
			-- ignored
			Card_Type := '$';
			return;
		     end if;
		  end loop;
	       end if;
	       
	       if R(0) > 0 then
		  if R2(0) > 0 then
		     J := R2(1) + 1;
		  else
		     J := R(1) + 1;
		  end if;
	       else
		  J := Mnemo(1)+1;
	       end if;
	       
	       while Is_White(Incard(J)) and J<=Inlen loop
		  J := J+1;
	       end loop;
	       
	       if J<=Inlen then
		  Arg(0) := J;
		  while J<=Inlen and then (not Is_White(Incard(J)) or Instring)  loop
		     if Incard(J)=''' then
			Instring := not Instring;
		     end if;
		     J := J+1;
		  end loop;
		  Arg(1) := J-1;
		  Parse_Arg;
	       end if;
	    end if;
	    if Mnemo(0)>0 then
	       exit;
	    end if;
	 end loop;
      end if;
   end Parse_Card;
   
   function Dprintf(I: Integer; N: Integer:= 42; Pad: Character:= '0') return String is
      
      Rv: String(1..N):= (others => Pad);
      J: Integer:= I;
      L: Integer:= 0;
   begin
      for K in reverse 1..N loop
	 Rv(K) := Character'Val(J mod 10 + Character'Pos('0'));
	 J := J/10;
	 L := L+1;
	 if J=0 then
	    exit;
	 end if;
      end loop;
      -- if j>0 then overflow
      if N=42 then -- 42 is also symbolic, it means 'any length' [but of course not larger than 42]
	 return "" & Rv(N-L+1..N);
      else
	 return Rv;
      end if;
   end Dprintf;
   
   function Isotime(T: Time) return String is
      
   begin
      return Dprintf(Year(T),4) & '-' &
	Dprintf(Month(T),2) & '-' &
	Dprintf(Day(T),2) & ';' &
	Dprintf(Integer(Long_Float'Floor(Long_Float(Seconds(T)))) / 3600,2) & ':' &
	Dprintf((Integer(Long_Float'Floor(Long_Float(Seconds(T)))) mod 3600)/60,2) & ':' &
	Dprintf(Integer(Long_Float'Floor(Long_Float(Seconds(T)))) mod 60,2) & '.' &
	Dprintf(Integer(1000.0 * (Long_Float(Seconds(T)) - Long_Float'Floor(Long_Float(Seconds(T))))),3)	
	;
   end Isotime;
   
   
   procedure Print_Job_Banner is
      
   begin	
      
      Put_Line("+------------------------------------------------------------------------------+");
      Put_Line("|  _____    _ _            ____ ____ _  __                                     |");
      Put_Line("| |  ___|__| (_)_  __     / ___/ ___/ |/ /_     RASSIRIS jobs simulator V0.3.0 |"); 
      Put_Line("| | |_ / _ \ | \ \/ /____| |  | |   | | '_ \    Job: " & Jn & "     by: " & Pn & " |");
      Put_Line("| |  _|  __/ | |>  <_____| |__| |___| | (_) |   Acn: " & An & "     RN: " & Runn & " |");
      Put_Line("| |_|  \___|_|_/_/\_\     \____\____|_|\___/    Start: " & Isotime(Clock) & " |");
      Put_Line("| http://dan.corlan.net/software/emulators/felix-assiris/                      |");
      Put_Line("+------------------------------------------------------------------------------+");
      New_Line;
      
   end Print_Job_Banner;
   
   procedure Print_Asm_Header is
      
   begin
      Put(Heading(1, "ASSEMBLER  LISTING", 1, 1));
      Put_Line("ADDR ASSEMBLY I B  Q OP X  D   LINE NR: LABEL     INSTRUCTION");
      New_Line;
   end Print_Asm_Header;
   
   procedure Print_Asm_Ender is
      
   begin
      Put_Line("ADDR ASSEMBLY I B  Q OP X  D   LINE NR: LABEL     INSTRUCTION");
      New_Line(2);
   end Print_Asm_Ender;
   
   procedure Print_Card(With_Assembly: Boolean) is
      Nr: String:= Integer'Image(Cardno);
      Snr: String(1..6) := ( others => ' ' );
      Slabel: String(1..10) := ( others => ' ' );
      Smnemor: String(1..10) := ( others => ' ');
      Sarg: String(1..64) := ( others => ' ');
      Sasm: String(1..18) := ( others => ' ');
      Sd: String := Half'Image(Ins.D);
      Si: Card;
      
   begin
      Snr(Snr'Last-Nr'Length+1..Snr'Last) := Nr;
      
      if Verbosity > 40 then
	 Put(Nr & "/" & Card_Type & "/");
	 if Label(0)>0 then
	    Put(Incard(Label(0)..Label(1)));
	 end if;
	 Put("/");
	 if Mnemo(0)>0 then
	    Put(Incard(Mnemo(0)..Mnemo(1)));
	 end if;
	 Put("/");
	 if R(0)>0 then
	    Put(Incard(R(0)..R(1)));
	 end if;
	 Put("/");
	 if Arg(0)>0 then
	    Put(Incard(Arg(0)..Arg(1)));
	 end if;
	 New_Line;
      end if;
      
      if With_Assembly then
	 Sasm(1) := Iflag'Image(Ins.I)(2);
	 Sasm(3) := Bfld'Image(Ins.B)(2);
	 if Ins.Q > 9 then
	    Sasm(5..6) := Qfld'Image(Ins.Q)(2..3);
	 else
	    Sasm(6) := Qfld'Image(Ins.Q)(2);
	    Sasm(5) := '0';
	 end if;
	 Sasm(8) := To_Hex_Digit(Natural(Ins.Op) / 16);
	 Sasm(9) := To_Hex_Digit(Natural(Ins.Op) mod 16);
	 Sasm(11) := Xflag'Image(Ins.X)(2);
	 Sasm(13..12+Sd'Length) := Sd;
      else
	 null;
      end if;
      
      if Verbosity>40 and Nextarg>1 then
	 -- print the parsed arguments
	 for K in 1..Nextarg-1 loop
	    if Args(K).Name(0)>0 then
	       Put(Incard(Args(K).Name(0)..Args(K).Name(1)) & ": ");
	    end if;
	    case Args(K).T is
	       when Longstr =>
		  Put_Line("S/" & Args(K).Sval(1..Args(K).Len) & "/");
	       when Ch =>
		  Si(1) := Character'Val(Args(K).Ival / 256);
		  Si(2) := Character'Val(Args(K).Ival mod 256);
		  if Si(1)<' ' or Si(1)>'~'then
		     Si(1) := '?';
		  end if;
		  if Si(2)<' ' or Si(2)>'~' then
		     Si(2) := '?';
		  end if;
		  Put_Line("C/" & Integer'Image(Args(K).Ival) & 
			     "/" & Si(1..2));
	       when Int =>
		  Put_Line("I/" & Integer'Image(Args(K).Ival));
	       when others =>
		  Put_Line("???");
	    end case;
	 end loop;
      end if;
      
      Put(To_Hex_Half(Card_Addr));
      Put(" ");
      if With_Assembly then
	 Put(To_Hex_Word(Inw));
      else
	 Put("        ");
      end if;
      Put(" ");
      Put(Sasm);
      
      if Label(0) > 0 then
	 Slabel(Slabel'First..Slabel'First+Label(1)-Label(0)) := "" & Incard(Label(0)..Label(1));
      end if;
      if Mnemo(0)>0 then
	 if R(0)>0 then
	    if R2(0)>0 then
	       declare
		  Mnemor: String:= "" & Incard(Mnemo(0)..Mnemo(1)) & "," & Incard(R(0)..R(1))
		    & "," & Incard(R2(0)..R2(1));
	       begin
		  Smnemor(Smnemor'First..Smnemor'First+Mnemor'Last-Mnemor'First) := Mnemor;
	       end;	       
	    else
	       declare
		  Mnemor: String:= "" & Incard(Mnemo(0)..Mnemo(1)) & "," & Incard(R(0)..R(1));
	       begin
		  Smnemor(Smnemor'First..Smnemor'First+Mnemor'Last-Mnemor'First) := Mnemor;
	       end;
	    end if;
	 else
	    Smnemor(Smnemor'First..Smnemor'First+Mnemo(1)-Mnemo(0)) := Incard(Mnemo(0)..Mnemo(1));
	 end if;
      end if;
      if Arg(0)>0 then
	 Sarg(Sarg'First..Sarg'First+Arg(1)-Arg(0)) := Incard(Arg(0)..Arg(1));
      end if;
      Put_Line(Snr & ": " & Slabel & Smnemor & Sarg);
   end Print_Card;
   
   function Heading(N: Natural; Title: String; Pre: Natural:= 0; Post: Natural := 0) return String is
      
      function Nchars(N: Natural; Char: Character) return String is
	 
      begin
	 if N=0 then
	    return "";
	 else
	    declare 
	       Rv: String(1..N) := ( others => Char );
	    begin
	       return Rv;
	    end;
	 end if;
      end Nchars;
      
   begin
      case N is
	 when 1 => 
	    return Nchars(Pre, Ascii.Lf) & Nchars(Title'Length+16, '*') & Ascii.Lf & "*** *** " & Title &
	      " *** ***" & Ascii.Lf & Nchars(Title'Length+16, '*') & Ascii.Lf & Nchars(Post, Ascii.Lf);
	 when 2 =>
	    return Nchars(Pre, Ascii.Lf) & Nchars(Title'Length+4, '=') & 
	      Ascii.Lf & "= " & Title & " =" & Ascii.Lf & 
	      Nchars(Title'Length+4, '=') & Ascii.Lf & Nchars(Post, Ascii.Lf);
	 when 3 =>
	    return Nchars(Pre, Ascii.Lf) &  Title & Ascii.Lf & 
	      Nchars(Title'Length, '-') & Ascii.Lf & Nchars(Post, Ascii.Lf);
	 when others =>
	    return Title;
      end case;
      
   end Heading;
   
   procedure Print_About is
      
   begin 
      Put(Heading(1, "About FELIX/ASSIRIS",3,1));
      Put_Line("FELIX V0.2.0., a felix/siris-2/assiris emulator, part of rassiris project.");
      Put_Line("see http://dan.corlan.net/software/emulators/felix-assiris/ for details.");
      Put_Line("");
      Put_Line("FELIX-C was a series of computers produced in Romania in the 1970-s");
      Put_Line("that were clones, more or less, of the french IRIS-50, that were");
      Put_Line("modified clones of the SDS Sigma series computers.");
      Put_Line("SIRIS was the operating system and ASSIRIS the assembly language.");
      New_Line;
      Put_Line("This is an approximate simulator for ASSIRIS jobs on the Felix.");
      Put_Line("It was written by Alexandru Dan Corlan and made available under");      
      Put_Line("the GPL v2 licence. ");      
      New_Line;
      Put_Line("Only a subset of the instructions are currently implemented. ");
      Put_Line("A few instructions are added to connect to the modern operating");
      Put_Line("system (Linux) under which the simulator runs. Some control cards and"); 
      Put_Line("directives are implemented, a few more are added.");
      New_Line;
      Put_Line("The felix command accepts no arguments, all instructions must be");
      Put_Line("given on standard input, that represents the card reader. Results");
      Put_Line("will be printed on the standard output that represents the system printer.");
      New_Line;
      Put_Line("More help is available by submitting (on stdin) a simple job, listed below:");
      New_Line;
      Put_Line(". LIST 'HELP'");      
      New_Line;
      Put(Heading(2, "Code of Conduct",1,1));
      Put_Line("  1.- Nothing may work, or work as expected, there is no warranty whatsoever, stay calm.");
      Put_Line("  2.- Any resemblance to an existing or historical computer system is purely coincidental");
      Put_Line("      (althought we tried, a little); don't hope.");
      Put_Line("  3.- Doing this is a complete, 100%, waste of time (we can proove that); relax.");
      Put_Line("  4.- Total utility of this project and of what you are doing right now is −273.15 Celsius (0K, Absolute Zero).");
   end Print_About;
   

   procedure Print_Help is
      
   begin
      Put(Heading(1, "FELIX/ASSIRIS VIRTUAL PROCESSOR HELP", 3, 1));
      Put(Heading(2, "GENERALITIES", 0, 1));
      Put_Line("FELIX V0.3.0., a felix/siris-2/assiris emulator, part of rassiris project.");
      Put_Line("see http://dan.corlan.net/software/emulators/felix-assiris/ for details.");
      Put_Line("");
      Put_Line("This version only has 64K of memory. It implements the following instructions:");
      Put_Line("     AD4 AD4I AD8 ADF8 ADF4 ADH2 BRU BCF BCT BAL CP1I CP1 CP2 CP4 CP8 CSV");
      Put_Line("     DC4 DV2 DVU2 DV4 DVF8 DVF4 EO2 EO4 EX2 EX4 IC2 IC4 LDC2 LDC4 LD1I LD2I");
      Put_Line("     LD1 LDH2 LDL2 LD4 LD8 LDM LD4I MG2 MG4 MP2 MPU2 MP4 MPF8 MPF4 NF4 SB4I");
      Put_Line("     SB4 SB8 SBH2 SH2 SH4 SH8 SBF8 SBF4 ST1 ST4 ST8 STH2 STM");
      Put_Line("to which we added a couple more (see below): PRINT HALT CSV HOST");
      Put_Line("The following pseudoinstructions are also implemented:");
      Put_Line("     BC BGEZ BGZ BLEZ BLZ BNC BNOV BNZ BOV BZ NOP SLL2 SLL4 SLL8 SRL2 SRL4");
      Put_Line("     SRL8 SLC2 SLC4 SLC8 SRC2 SRC4 SRC8 SRA2 SRA4 SRA8 SLA2 SLA4 SLA8 ASCZ");
      Put_Line("     ASCNZ ASCC ASCNC ASCS ASCNS ASCD ASCND ASCT ASCF ASEQ TERM ABORT");
      Put_Line("All addressing modes are implemented in this version, direct-nonindexed, indirect-nonindexed,");
      Put_Line("    direct-indexed and indirect-indexed.");
      Put_Line("Indirect addressing performs exactly one indirection.");
      Put(Heading(2, "CONTROL CARDS", 1, 1));
      Put_Line("The following cards (LIST and CONF) can appear anywhere in the input stream:");
      Put_Line(". LIST opt{,opt}* where opt can be:");
      Put_Line("       'HELP'     -- include this help text in the listing");
      Put_Line("       'ABOUT'    -- introduction to the FELIX/ASSIRIS system");
      Put_Line("       'SYMS'     -- the symbols table");
      Put_Line("       'LINKS'    -- the linkings performed by the link editor up to that point");
      Put_Line("       'DUMP'     -- `VIDAGE MEMOIRE' at that point");
      Put_Line("        MSG:'x'    -- display the message at that point");
      Put_Line(". JOB name,opt{,opt}*");
      Put_Line(". CONF opt{,opt}*");
      Put_Line("       JOB starts a job with a job name (max 8 characters) and options. CONF may change or provide the same options");
      Put_Line("       at an earlier or latter time. Options can be: AN=cccccccc account name, PN=cccccccc");
      Put_Line("       programmer name, VERB=nn verbosity level--0=no system message; 10=basic messaging, incl failed assertions");
      Put_Line("       20=extensive messages, incl passed assertions; 40=debugging messages., SPEED=N/O or 1/0 N or 1=native speed");
      Put_Line("       (default, as fast as possible), O or 0=original timing, 3--30 microseconds per instruction.");
      New_Line;
      Put_Line("The following cards can only appear in a specifc sequence.");
      Put_Line("The sequence is: JOB, COMPILE, LINK, RUN, EOJ; then, you may repeat.");
      Put_Line("COMPILE must be follwed by ASSIRIS (. COMPILE ASSIRIS)");
      Put_Line("RUN will admit one option, KINS:n where n is the number of thousands of");
      Put_Line("  instructions to run, or INS:n, where n is the number of instructions.");
      Put_Line("  The default is INS:500.");
      Put_Line("Otherwise, you may add any options to the cards, but they are currently ignored.");
      Put(Heading(2, "DIRECTIVES", 1, 1));
      Put_Line("Directives are cards that can occur only between '. COMPILE ASSIRIS' and 'END'");
      Put_Line("The directives: ORG, EQU, DS, DB, ALIGN, work as expected. EQU defines a symbol.");
      Put_Line("ORG changes the address (that must be its argument) where the assembler generates code");
      Put_Line("DATA,x,y datum{,datum}* stores each datum in succesive locations of x bytes each and ");
      Put_Line("  aligned at y; by default, x and y are 4. Each datum has the syntax mentioned at DB");
      Put_Line("  below, to which FS:'<float>' and FL:'<float>' are added and, of course, the");
      Put_Line("  size of the objects specified can be of the size x.");
      Put_Line("DS is followed by a '-delimited string, the ASCII characters of which it puts into memory");
      Put_Line("  in succesive byte locations.");
      Put_Line("TEXT is like DS.");
      Put_Line("TEXTC is like TEXT, but the string is preceded by a byte containing its length.");
      Put_Line("DB is followed by a sequence of byte-sized numbers (0..255),");
      Put_Line("  Comma separated, that ar put into memory in sequence.");
      Put_Line("  the number can be: [-]ddd, decimal numbers; X'xx' hexadecimal, C'c' characters.");
      Put_Line("ALIGN and BOUND (synonymous). Before assemblying machine code, ");
      Put_Line("  the assembly address must be aligned to 4 bytes.");
      Put_Line("  After DSs and DBs, always use ALIGN 4, if code follows.");
      Put_Line("  ALIGN take an optional argument--at what pace to align, in bytes. The default is 4.");
      Put_Line("END X must include this X which is the address, usually a label, where RUN will start execution.");
      Put_Line("CSECT is ignored.");
      Put_Line("Label expressions are not yet implemented, you can't say 'BRU ADDR+8' or something");
      Put(Heading(2, "NEW INSTRUCTIONS", 1, 1));
      Put_Line("HALT will finish running, but it is better to invoke it through TERM.");
      Put_Line("PRINT,r will print the LSB of register r as an ascii character on stdout");
      Put_Line("PRINT,r C'd' will print the register r as a decimal on stdout");
      Put_Line("PRINT,r C'xy' will print the two ASCII characters x and y on stdout");
      Put_Line("CSV and HOST are the same instruction (opcode) with different flags; do not");
      Put_Line("    use them individually, in general, use macros/pseudoinstructions");
      Put_Line("ASEQ,r val will check the assertion that general register r has the value val; it generates");
      Put_Line("    one HOST instruction that uses information (val) stored in a monitor table.");
      Put_Line("    When the host instruction is executed, it checks the value and produces a");
      Put_Line("    line in the output, depending on verbosity; a statistic of passed and failed");
      Put_Line("    assertions is also produced at the end with appropriate verbosity flags.");
      Put_Line("ASCZ ASCNZ ASCC ASCNC ASCS ASCNS ASCD ASCND ASCT ASCF ASEQ will check the condition");
      Put_Line("    and produce the appropriate reports depending on the verbosity level.");
   end Print_Help;
   
   -- THE ASSEMBLER
   
   
   Dir: Boolean:= False; -- if it was identified as a directive card
   
   Type Ops is (Ad4, Ad4i, 
		AD8, ADD, ADF8, ADF4, 
		Adh2, Adl2, 
		AIO, ANALD, ANALN,
		Bal, Bcf, Bct, Bdf, Bdt, Bru, Cp1, Cp1i, Cp2, Cp4, 
		BSG,
		CP8, CPD, CPSL, CPSR, CYBL, CYBR,
		CSV,
		Dv2, Dv4,
		DVD, DVF8, DVF4, DVU2, EO2, EO4, EX2, EX4, EXU, HIO,
		Ic2, Ic4, 
		LAS,
		Ld1, Ld1i, Ld2i, Ldh2, Ldl2, Ld4, 
		LD4I, LD8, LDC, LDC2, LDC4, LDC8, LDDA, LDM, LDS2, LDTM, LPK, LPS, 
		Mp2, Mp4, 
		MG2, MG4, MPD, MPF4, MPF8, MPU2, MVSR, MVSL,
		NF4, PACK, 
		Rd, 
		Sb4, Sb4i,
		SB8, SBD, SBF4, SBF8, SBH2,
		Sh2, Sh4, 
		SH8, SHD, SIO,
		St1, St4, Sth2, Stpa, 
		ST8, STC, STM, STS2, STTM,
		TDV, TIO, TRTL, TRTR, UNPK,
		Wd, 
		WT, ZAD,
		
		-- Felix-C-1024 extensions
		
		LDL1, STL1, STL2, ADM4, DC4, MPI4, DVI4, 
		EX1I, MG1I, EO1I, 
		-- NF4, ? exista deja
		LDCH, STCH, CVDB, CVBD, EDIT,
		
		
		-- added SEMIHOSTING instructions
		Print, Halt, 
		Host, -- not to be used anymore
		
		-- for the bcd use: https://gcc.gnu.org/onlinedocs/gcc-7.1.0/gccint/Libgcc.html#Libgcc
		
		UNASSIGNED);
   
   Assert_Hostop: constant Half:= 16#D0F0#;
   
   Byopcod: array(Opcod) of Ops := (others => UNASSIGNED);
   
   Byops: array(Ops) of Opcod := (others => 0); -- impossible opcode
   
   type Machines is ( FC256, FC1024, FCC16, FCC24, FCC31, FCX32, FCX64, None );
   
   Timings: array(Ops) of Duration;
   Machina: array(Ops) of Machines;
   
   Us: Duration:= 1.0e-6; -- microsecond
   
   procedure Setops(Op: Ops; 
		    Cod: Opcod; 
		    Machine: Machines := FC256; 
		    Timing: Long_Float:= 3.0 -- microsecond
		   ) is
      
   begin
      Byopcod(Cod) := Op;
      Timings(Op) := Duration(Timing) * Us;
      Machina(Op) := Machine;
      if Byops(Op) /= 0 then
	 Put_Line("Standard_Error: Code already exists: " & To_Hex_Half(Half(Cod)));
      else
	 Byops(Op) := Cod;
      end if;
   end Setops;
   
   procedure Set_All_Ops is
      
   begin
      Setops(Ad4, 16#1E#, Timing => 4.6);
      Setops(Ad4i, 16#27#, Timing => 4.3);
      Setops(AD8, 16#1F#, Timing => 7.35);
      Setops(ADD, 16#61#, Timing => 10.3 + 1.6 * 32.0 + 4.2 * 32.0);
      Setops(ADF8, 16#7C#, Timing => 9.45);
      Setops(ADF4, 16#6C#, Timing => 8.10);
      Setops(Adh2, 16#1D#, Timing => 3.6);
      Setops(Adl2, 16#1C#, Timing => 4.45);
      Setops(AIO, 16#56#, Timing => 19.95);
      Setops(ANALD, 16#3D#, Timing => 6.2);
      Setops(ANALN, 16#3C#, Timing => 6.2);
      Setops(Bal, 16#39#, Timing => 2.85);
      Setops(Bcf, 16#34#, Timing => 4.1);
      Setops(Bct, 16#35#, Timing => 4.1);
      Setops(Bdf, 16#36#, Timing => 4.1);
      Setops(Bdt, 16#37#, Timing => 4.1);
      Setops(Bru, 16#38#, Timing => 2.85);
      Setops(Bsg, 16#2F#);
      Setops(Cp1, 16#10#, Timing => 3.60);
      Setops(Cp1i, 16#24#, Timing => 2.85);
      Setops(Cp2, 16#11#, Timing => 3.60);
      Setops(Cp4, 16#12#, Timing => 4.60);
      Setops(CP8, 16#13#, Timing => 7.45);
      Setops(CPD, 16#63#, Timing => 9.05 + 1.6 * 32.0 + 2.6 * 32.0);
      Setops(CPSL, 16#4C#, Timing => 8.0);
      Setops(CPSR, 16#4D#, Timing => 8.0);
      Setops(Csv, 16#2E#);
      Setops(CYBL, 16#4A#, Timing => 6.75);
      Setops(CYBR, 16#4B#, Timing => 6.75);
      Setops(Dv2, 16#32#, Timing => 10.3);
      Setops(Dv4, 16#68#, Timing => 32.1);
      Setops(DVD, 16#65#, Timing => 8.45 + 1.6 * 32.0 + 3.2 * 32.0 + (1.6 + 0.95 * 32.0) * 32.0);
      Setops(DVF8, 16#7F#, Timing => 55.3);
      Setops(DVF4, 16#6F#, Timing => 27.3);
      Setops(DVU2, 16#33#, Timing => 10.35);
      Setops(EO2, 16#0D#, Timing => 3.6);
      Setops(EO4, 16#0E#, Timing => 4.6);
      Setops(EX2, 16#01#, Timing => 3.6);
      Setops(EX4, 16#02#, Timing => 4.6);
      Setops(EXU, 16#3A#, Timing => 2.85);
      Setops(HIO, 16#53#, Timing => 20.9);
      Setops(Ic2, 16#46#, Timing => 4.7);
      Setops(Ic4, 16#47#, Timing => 6.35);
      Setops(Las, 16#0C#, Timing => 4.35);
      Setops(Ld1, 16#18#, Timing => 3.85);
      Setops(Ld1i, 16#28#, Timing => 2.85);
      Setops(Ld2i, 16#29#, Timing => 2.85);
      Setops(Ldh2, 16#08#, Timing => 3.6);
      Setops(Ldl2, 16#09#, Timing => 4.1);
      Setops(Ld4, 16#0A#, Timing => 4.35);
      Setops(LD4I, 16#2A#, Timing => 3.45 );
      Setops(LD8, 16#0B#, Timing => 7.1);
      Setops(LDC, 16#43#, Timing => 4.25); -- load conditions (not complement)
      Setops(LDC2, 16#19#, Timing => 3.60);
      Setops(LDC4, 16#1A#, Timing => 4.35);
      Setops(LDC8, 16#1B#, Timing => 7.20);
      Setops(LDDA, 16#2B#, Timing => 2.85);
      Setops(LDM, 16#2C#, Timing => 7.1);
      Setops(LDS2, 16#45#, Timing => 4.25);
      Setops(LDTM, 16#41#, Timing => 4.25);
      Setops(LPK, 16#5D#, Timing => 6.75);
      Setops(LPS, 16#5E#, Timing => 6.95);
      Setops(Mp2, 16#30#, Timing => 8.95);
      Setops(Mp4, 16#69#, Timing => 28.0);
      Setops(MG2, 16#05#, Timing => 3.60);
      Setops(MG4, 16#06#, Timing => 4.60);
      Setops(MPD, 16#64#, Timing => 8.45 + 1.6*32.0 + 3.2*32.0 + (1.6 + 0.95 * 32.0)*32.0 + 1.5);
      Setops(MPF4, 16#6E#, Timing => 24.4);
      Setops(MPF8, 16#7E#, Timing => 51.45);
      Setops(MPU2, 16#31#, Timing => 9.3);
      Setops(MVSR, 16#49#, Timing => 7.65);
      Setops(MVSL, 16#48#, Timing => 7.65);
      Setops(NF4, 16#22#, Timing => 9.3);
      Setops(PACK, 16#66#, Timing => 7.45 + 1.6*(32.0 + 16.0));
      Setops(Rd, 16#50#, Timing => 3.35);
      Setops(Sb4, 16#16#, Timing => 4.6);
      Setops(Sb4i, 16#26#, Timing => 4.3);
      Setops(SB8, 16#17#, Timing => 7.45);
      Setops(SBD, 16#62#, Timing => 10.3 + 1.6 * 32.0 + 4.2 * 32.0);
      Setops(SBF4, 16#6D#, Timing => 8.1);
      Setops(SBF8, 16#6F#, Timing => 9.45);
      Setops(SBH2, 16#15#, Timing => 3.60);
      Setops(Sh2, 16#20#, Timing => 3.85);
      Setops(Sh4, 16#21#, Timing => 3.85);
      Setops(SH8, 16#6B#, Timing => 44.75);
      Setops(SHD, 16#5F#, Timing => 8.3 + 3.2 * 32.0);
      Setops(SIO, 16#52#, Timing => 20.9);
      Setops(St1, 16#58#, Timing => 3.85); -- stl, 2?
      Setops(St4, 16#5A#, Timing => 4.60);
      Setops(Sth2, 16#59#, Timing => 3.60);
      Setops(Stpa, 16#5C#, Timing => 4.60); 
      Setops(ST8, 16#5B#, Timing => 6.60);
      Setops(STC, 16#42#, Timing => 7.35);
      Setops(STM, 16#2D#, Timing => 7.16);
      Setops(STS2, 16#44#, Timing => 7.35);
      Setops(STTM, 16#40#, Timing => 7.35);
      Setops(TDV, 16#55#, Timing => 20.9);
      Setops(TIO, 16#54#, Timing => 20.9);
      Setops(TRTL, 16#4E#, Timing => 11.57);
      Setops(TRTR, 16#4F#, Timing => 11.57);
      Setops(UNPK, 16#67#, Timing => 8.3 + 1.6*(32.0 + 16.0));
      Setops(Wd, 16#51#, Timing => 3.35);
      Setops(WT, 16#57#, Timing => 20.9);
      Setops(ZAD, 16#60#, Timing => 7.2 + 1.6*(32.0 + 32.0));
      
      if Show_Opcodes then
	 Put_Line(Standard_Error, "Unassigned opcodes in Felix-C-256: ");
	 for Op in Opcod loop
	    if Byopcod(Op)=Unassigned then
	       Put(To_Hex_Half(Half(Op)) & ", ");
	    end if;
	 end loop;
      end if;
      
      --    Unused codes in Felix-C-256:
      
      --    0000, 
      --    0003, 0004, 0007, 000C, 000F, 
      --    0014, 001B, 0023, 0025, 002E, 
      --    002F, 003B, 003E, 003F, 006A, 
      --    0070, 0071, 0072, 0073, 0074, 
      --    0075, 0076, 0077, 0078, 0079, 
      --    007A, 007B, 007D
      
      -- Felix C 1024 new instructions
      
      Setops(Ldl1, 16#0F#, Machine => Fc1024);
      Setops(Stl1, 16#70#, Machine => Fc1024);
      Setops(Stl2, 16#71#, Machine => Fc1024);
      Setops(Adm4, 16#77#, Machine => Fc1024);
      Setops(Dc4,  16#3F#, Machine => Fc1024);
      Setops(Mpi4, 16#04#, Machine => Fc1024);
      Setops(Dvi4, 16#14#, Machine => Fc1024);
      Setops(Ex1i, 16#03#, Machine => Fc1024);
      Setops(Mg1i, 16#07#, Machine => Fc1024);
      --    Setops(Eo1i, 16#26#, Machine => Fc1024); e deja sb4i !
      --    Setops(Nf4,  16#6A#, Machine => Fc1024);
      Setops(Ldch, 16#73#, Machine => Fc1024);
      Setops(Stch, 16#72#, Machine => Fc1024);
      Setops(Cvdb, 16#75#, Machine => Fc1024); -- 23 on Felix C512
      Setops(Cvbd, 16#74#, Machine => Fc1024); -- 3E on Felix C512
      Setops(Edit, 16#3B#, Machine => Fc1024);
      
      if Show_Opcodes then
	 New_Line;
	 Put_Line(Standard_Error, "Unassigned opcodes in Felix-C-1024: ");
	 for Op in Opcod loop
	    if Byopcod(Op)=Unassigned then
	       Put(To_Hex_Half(Half(Op)) & ", ");
	    end if;
	 end loop;
      end if;
      
      
      -- Felix-CC-16 simulator instructions
      
      Setops(Print,16#78#, Machine => Fcc16);
      Setops(Halt,16#79#, Machine => Fcc16);
      Setops(Host,16#7A#, Machine => Fcc16);
      -- mai este si 7B liber
      
      if Show_Opcodes then
	 New_Line;
	 Put_Line(Standard_Error, "Unassigned opcodes in Felix-CC16: ");
	 for Op in Opcod loop
	    if Byopcod(Op)=Unassigned then
	       Put(To_Hex_Half(Half(Op)) & ", ");
	    end if;
	 end loop;
      end if;
      
      
   end Set_All_Ops;
   
   
   procedure Asminc(Op: Opcod; Q: Qfld; B: Bfld; D: Half; Ii: Boolean; Xi: Boolean) is
      
   begin
      while Asm mod 4 /= 0 loop -- automatic bound (align) 4
	 Asm := Asm+1;
      end loop;
      Ins.Op := Op;
      Ins.D := D;
      Ins.Q := Q;
      if Xi then
	 Ins.X := 1;
      else
	 Ins.X := 0;
      end if;
      Ins.B := B;
      if Ii then
	 Ins.I := 1;
      else
	 Ins.I := 0;
      end if;
      M(Asm..Asm+3) := Inb;
      Asm := Asm+4;
   end Asminc;
   
   
   procedure Assemble_Ins(Opid: String; Rarg0: Natural; Barg0: Bfld; Disp0: Natural;
								     X,I: Boolean:= False) is
      
      Rarg: Qfld:= Qfld(Rarg0);
      Disp: Half:= Half(Disp0);
      
      procedure Assemble_Condition_Assert(Ast: Assertion_Type) is
	 
      begin
	 Asr(Next_Asr).Adr := Asm;
	 Asr(Next_asr).Reg := Qfld(Rarg0);
	 Asr(Next_Asr).Assertt := Ast;
	 Next_Asr := Next_Asr+1;
	 Asminc(Byops(Csv),Rarg,0,Assert_Hostop,False,True);	 
      end Assemble_Condition_Assert;

   begin
      while (Asm mod 4) /= 0 loop -- always align instructions
	 Asm := Asm + 1;
      end loop;
      for K in Ops loop
	 if Opid=Ops'Image(K) then
	    Asminc(Byops(K),Rarg,Barg0,Disp,I,X);
	    return;
	 end if;
      end loop;
      
      -- pseudo-ops and macros
      
      -- branch pseudo-ops
      
      if Opid="BC" then
	 Asminc(Byops(Bct),1,Barg0,Disp,I,X);
	 return;
      elsif Opid="BGEZ" then
	 Asminc(Byops(Bcf),4,Barg0,Disp,I,X);
	 return;
      elsif Opid="BGZ" then
	 Asminc(Byops(Bcf),12,Barg0,Disp,I,X);
	 return;
      elsif Opid="BLEZ" then
	 Asminc(Byops(Bct),12,Barg0,Disp,I,X);
	 return;
      elsif Opid="BLZ" then
	 Asminc(Byops(Bct),4,Barg0,Disp,I,X);
	 return;
      elsif Opid="BNC" then
	 Asminc(Byops(Bcf),1,Barg0,Disp,I,X);
	 return;
      elsif Opid="BNOV" then
	 Asminc(Byops(Bcf),2,Barg0,Disp,I,X);
	 return;
      elsif Opid="BNZ" then
	 Asminc(Byops(Bcf),8,Barg0,Disp,I,X);
	 return;
      elsif Opid="BOV" then
	 Asminc(Byops(Bct),2,Barg0,Disp,I,X);
	 return;
      elsif Opid="BZ" then
	 Asminc(Byops(Bct),8,Barg0,Disp,I,X);
	 return;
      elsif Opid="NOP" then
	 Asminc(Byops(Bct),0,Barg0,Disp,I,X);
	 return;
	 
	 --- shift pseudoops
	 
      elsif Opid="SLL2" then
	 Asminc(Byops(Sh2),Rarg,0,0+Disp,FALSE,FALSE);
	 return;
      elsif Opid="SLL4" then
	 Asminc(Byops(Sh4),Rarg,0,0+Disp,FALSE,FALSE);
	 return;
      elsif Opid="SLL8" then
	 Asminc(Byops(Sh8),Rarg,0,0+Disp,FALSE,FALSE);
	 return;
      elsif Opid="SRL2" then
	 Asminc(Byops(Sh2),Rarg,0,16#400#+Disp,FALSE,FALSE);
	 return;
      elsif Opid="SRL4" then
	 Asminc(Byops(Sh4),Rarg,0,16#400#+Disp,FALSE,FALSE);
	 return;
      elsif Opid="SRL8" then
	 Asminc(Byops(Sh8),Rarg,0,16#400#+Disp,FALSE,FALSE);
	 return;
      elsif Opid="SLC2" then
	 Asminc(Byops(Sh2),Rarg,0,16#100#+Disp,FALSE,FALSE);
	 return;
      elsif Opid="SLC4" then
	 Asminc(Byops(Sh4),Rarg,0,16#100#+Disp,FALSE,FALSE);
	 return;
      elsif Opid="SLC8" then
	 Asminc(Byops(Sh8),Rarg,0,16#100#+Disp,FALSE,FALSE);
	 return;
      elsif Opid="SRC2" then
	 Asminc(Byops(Sh2),Rarg,0,16#500#+Disp,FALSE,FALSE);
	 return;
      elsif Opid="SRC4" then
	 Asminc(Byops(Sh4),Rarg,0,16#500#+Disp,FALSE,FALSE);
	 return;
      elsif Opid="SRC8" then
	 Asminc(Byops(Sh8),Rarg,0,16#500#+Disp,FALSE,FALSE);
	 return;
      elsif Opid="SRA2" then
	 Asminc(Byops(Sh2),Rarg,0,16#600#+Disp,FALSE,FALSE);
	 return;
      elsif Opid="SRA4" then
	 Asminc(Byops(Sh4),Rarg,0,16#600#+Disp,FALSE,FALSE);
	 return;
      elsif Opid="SRA8" then
	 Asminc(Byops(Sh8),Rarg,0,16#600#+Disp,FALSE,FALSE);
	 return;
      elsif Opid="SLA2" then
	 Asminc(Byops(Sh2),Rarg,0,16#200#+Disp,FALSE,FALSE);
	 return;
      elsif Opid="SLA4" then
	 Asminc(Byops(Sh4),Rarg,0,16#200#+Disp,FALSE,FALSE);
	 return;
      elsif Opid="SLA8" then
	 Asminc(Byops(Sh8),Rarg,0,16#200#+Disp,FALSE,FALSE);
	 return;
	 -- assertions for conditions
      elsif Opid="ASCZ" then
	 Assemble_Condition_Assert(Z);
      elsif Opid="ASCNZ" then
	 Assemble_Condition_Assert(NZ);
      elsif Opid="ASCC" then
	 Assemble_Condition_Assert(C);
      elsif Opid="ASCNC" then
	 Assemble_Condition_Assert(NC);	
      elsif Opid="ASCS" then
	 Assemble_Condition_Assert(S);
      elsif Opid="ASCNS" then
	 Assemble_Condition_Assert(NS);
      elsif Opid="ASCD" then
	 Assemble_Condition_Assert(D);
      elsif Opid="ASCND" then
	 Assemble_Condition_Assert(ND);
      elsif Opid="ASCT" then
	 Asr(Next_Asr).Ival := Args(1).Ival;
	 Assemble_Condition_Assert(CT);
      elsif Opid="ASCF" then
	 Asr(Next_Asr).Ival := Args(1).Ival;
	 Assemble_Condition_Assert(CF);
	 -- assertions for register value
      elsif Opid="ASEQ" then      --- BUG! when not an instruction, the following comment is parsed, not ignored
	 Asr(Next_Asr).Adr := Asm;
	 Asr(Next_asr).Reg := Qfld(Rarg0);
	 case Args(1).T is
	    when Int =>
	       Asr(Next_Asr).Assertt := Int_Word_Eq;
	       Asr(Next_Asr).Ival := Args(1).Ival;
	    when Fx =>
	       Asr(Next_Asr).Assertt := Float_Eq;
	       Asr(Next_Asr).Fval := Args(1).Fval;
	    when Dx =>
	       Asr(Next_Asr).Assertt := Double_Eq;
	       Asr(Next_Asr).Dval := Args(1).Dval;
	    when others =>
	       Asm_Error("Wrong assertion type.");
	 end case;
	 Next_Asr := Next_Asr+1;
	 Asminc(Byops(Csv),Rarg,0,Assert_Hostop,False,True);
	 return;
      end if;
      
      Put_Line("Invalid opcode: " & Opid);
   end Assemble_Ins;
   
   procedure Assemble_Card is 
      -- card that has just been read and parsed, found in incard
      -- and parsed in mnemo & folks
      
      
      function Reg_Arg return Bfld is
	 
      begin
	 if Nextarg < 2 then
	    return 0;
	 end if;
	 if Args(1).T=Reg then
	    if Args(1).Ival > 0 and then Args(1).Ival<9 then
	       return Bfld(Args(1).Ival);
	    elsif
	      Args(1).Ival > 8 and then Args(1).Ival < 16 then
	       return Bfld(Args(1).Ival - 8);
	    else
	       Asm_Error("Indirect, B register must be in 1..8 or 9..16, 1..8 meaning 9..16.");
	    end if;
	 end if;
	 return 0;
      end Reg_Arg;
      
      function Half_Arg return Half is
	 
      begin
	 if Nextarg < 2 then
	    return 0;
	 end if;
	 if Args(1).T = Reg then
	    if Nextarg=3 then
	       return Args(2).Hval;
	    else
	       return 0;
	    end if;
	 else
	    return Args(1).Hval;
	 end if;
      end Half_Arg;
      
      Rr, Rb: Integer;
      Disp: Integer;
      Ifl, Xfl: Boolean;
      Argerr: Boolean;
      
      procedure Parse_Instruction_Args is
	 
	 -- returns the above, supersedes parse_args for instructions
	 
      begin
	 Argerr := False;
	 Ifl := False;
	 Xfl := False;
	 Disp := - (2**17);  -- impossible value
	 Rr := -1; -- impossible value
	 Rb := -1; -- impossible value
	 
	 if Nextarg>3 then -- this is not an instruction
	    Argerr := True;
	    return;
	 end if;
	 
	 if Nextarg<2 then -- there is no argument, this is ok
	    Disp := 0;
	    Rr := 0;
	    Rb := 0;
	    return;
	 end if;
	 
	 declare 
	    Arg1: String := "" & Incard(Args(1).Val(0)..Args(1).Val(1));
	    F1: Integer:= Arg1'First; -- cursor in arg1
	 begin
	    if Arg1="*" then
	       Disp := Integer(Asm);
	       Rr := 0;
	       Rb := 0;
	       return;
	    end if;
	    if Arg1="%" then
	       -- only in FCC16
	       Disp := 0;
	       Rr := 0;
	       Rb := 0;
	       return;
	    end if;
	    if Arg1(F1)='*' then
	       -- it is 'starred', so indirect addressing
	       Ifl := True;
	       F1 := F1+1;
	    end if;
	    if Arg1'Last-F1 > 2 and then Arg1(F1)='B' and then Arg1(F1+1)>='0' and then Arg1(F1+1)<='7'
	      and then Arg1(F1+2)='.' then
	       Rb := Character'Pos(Arg1(F1+1)) - Character'Pos('0'); -- a Bn. syntax is present
	       F1 := F1+3;
	    else
	       Rb := 0;
	    end if;
	    -- now, we deal with the rest as being an Rn special label or a label or a constant	
	    --	    Put_Line("before disp");
	    Disp := Integer(To_Half("" & Arg1(F1..Arg1'Last))); -- that's where the detailed analysis should be	
								--	    Put_Line("after disp");    
	    if Nextarg=3 then
	       Xfl := True;
	       -- this must resolve to 0..15
	       declare
		  Arg2: String := "" & Incard(Args(2).Val(0)..Args(2).Val(1));
	       begin
		  if Arg2(Arg2'First)='R' and then Arg2'Length>1 and then
		    Is_Declit(Arg2(Arg2'First+1..Arg2'Last)) then
		     Rr := Integer'Value(Arg2(Arg2'First+1..Arg2'Last));
		  elsif Is_Declit(Arg2) then
		     Rr := Integer'Value(Arg2);
		  else
		     Argerr := True;
		  end if;
		  if Rr<0 or else Rr>15 then
		     Argerr := True;
		     Rr := -1;
		  end if;
		  return;
	       end;
	    end if;
	 end;
	 return;
      end Parse_Instruction_Args;
      
      Alignat: Half := 4;
      
   begin
      Card_Addr := Asm;
      
      -- NEW LABELS
      if Label(0) > 0 then
	 if Mnemo(0) > 0 and then "" & Incard(Mnemo(0)..Mnemo(1)) = "EQU" then
	    Add_To_Table(Incard(Label(0)..Label(1)), Word(Half_Arg)); -- but must check for the sign!
	    return;
	 else
	    -- NU E BINE! Daca e instructiune/pseudo/macro, se genereaza ceva, atunci trebuie aliniat la 4!
	    -- daca sunt date nu trebuie aliniat
	    Add_To_Table(Incard(Label(0)..Label(1)), Word(Asm));
	 end if;
      end if;
      
      -- DIRECTIVES
      Dir := False;
      if Mnemo(0) > 0 then
	 if "RES" = "" & Incard(Mnemo(0)..Mnemo(1)) then
	    if Nextarg/=2 or else Args(1).T /= Int then
	       Asm_Error("RES needs a number of bytes to reserve.");
	    end if;
	    -- first, align
	    if R(0)>0 then 
	       Rr := Integer'Value(Incard(R(0)..R(1)));
	       while (Asm mod Half(Rr)) /= 0 loop
		  Asm := Asm + 1;
	       end loop;
	       Asm := Asm + Half(Args(1).Ival * Rr);
	    else
	       Asm := Asm + Half(Args(1).Ival);
	    end if;
	    Dir := True;
	 elsif "ALIGN" = "" & Incard(Mnemo(0)..Mnemo(1)) or else 
	   "BOUND" = "" & Incard(Mnemo(0)..Mnemo(1)) then
	    if Nextarg=2 and Args(1).T = Int then
	       Alignat := Half(Args(1).Ival);
	    end if;
	    while (Asm mod Alignat) /= 0 loop
	       Asm := Asm + 1;
	    end loop;
	    Dir := True;
	 elsif "DS" = Incard(Mnemo(0)..Mnemo(1)) or else
	   "TEXT" = Incard(Mnemo(0)..Mnemo(1)) or else 
	   "TEXTC" = Incard(Mnemo(0)..Mnemo(1))then
	    if Nextarg/=2 or else Args(1).T /= Str then
	       Asm_Error("TEXT/TEXTC/DS needs a string to store.");
	    end if;
	    if Incard(Mnemo(0)..Mnemo(1)) = "TEXTC" then
	       if Args(1).Len > 255 then
		  Asm_Error("TEXTC admits only strings up to 255 characters.");
	       end if;
	       M(Asm) := Byte(Args(1).Len);
	       Asm := Asm+1;
	    end if;
	    for K in 1..Args(1).Len loop
	       M(Asm) := Byte(Character'Pos(Args(1).Sval(K)));
	       Asm := Asm+1;
	    end loop;
	    if Incard(Mnemo(0)..Mnemo(1)) = "DS" then
	       M(Asm) := 0;
	       Asm := Asm+1;
	    end if;
	    Dir := True;
	 elsif "ORG" = "" & Incard(Mnemo(0)..Mnemo(1)) then
	    if Nextarg/=2 or else Args(1).T /= Int then
	       Asm_Error("ORG needs an address from where to assemble.");
	    end if;
	    Asm := Half(Args(1).Ival);
	    Dir := True;
	 elsif "LIST" = "" & Incard(Mnemo(0)..Mnemo(1)) then
	    for K in 1..Nextarg-1 loop
	       if Args(K).Name(0) > 0 and then
		 Incard(Args(K).Name(0)..Args(K).Name(1)) = "VERB"
		 and then Args(K).T=Int  then
		  Set_Verbosity(Args(K).Ival);
	       elsif Args(K).T=Str and then Args(K).Sval(1..Args(K).Len) = "HELP" then
		  Print_Help;
	       elsif Args(K).T=Str and then Args(K).Sval(1..Args(K).Len) = "ABOUT" then
		  Print_Help;
	       end if;
	    end loop;
	    Dir := True;
	 elsif "END" = "" & Incard(Mnemo(0)..Mnemo(1)) then
	    if Nextarg/=2 or else Args(1).T /= Lbl then
	       Asm_Error("END needs an address from where the program will run.");
	    end if;
	    Run_Entry := Args(1).Hval;
	    Dir := True;
	    Incompile := False;
	 elsif "DATA" = "" & Incard(Mnemo(0)..Mnemo(1)) then
	    declare
	       Dsz, Dalign: Integer:= 4;
	    begin
	       if R(0)>0 then
		  Dsz := Integer'Value(Incard(R(0)..R(1)));
	       end if;
	       if R2(0)>0 then
		  Dalign := Integer'Value(Incard(R2(0)..R2(1)));
	       end if;
	       
	       -- first, align
	       
	       while Asm mod Half(Dalign) /= 0 loop
		  Asm := Asm + 1;
	       end loop;
	       
	       for K in 1..Nextarg-1 loop
		  if Args(K).T=Int then
		     case Dsz is
			when 1 =>
			   M(Asm) := Byte(Args(K).Ival mod 256);
			when 2 =>
			   Memstor2(Asm, Args(K).Ival);
			when 4 =>
			   Memstor4(Asm, Args(K).Ival);
			when 8 => -- this must be sorted somehow, it must be possible to specify a long and represent it in args
			   Memstor8(Asm, Long_Integer(Args(K).Ival));
			when others =>
			   Asm_Error("Unknown integer representation.");
		     end case;
		     Asm := Asm+Half(Dsz);
		  elsif Args(K).T=Fx then
		     if Dsz/=4 then
			Asm_Error("Wrong size of data statement for short float: " & Integer'Image(Dsz));
		     end if;
		     Memstor4(Asm, Args(K).Fval);
		     Asm := Asm+Half(Dsz);
		  elsif Args(K).T=Dx then
		     if Dsz/=8 then
			Asm_Error("Wrong size of data statement for long float: " & Integer'Image(Dsz));
		     end if;
		     Memstor8(Asm, Args(K).Dval);
		     Asm := Asm+Half(Dsz);
		  end if;
	       end loop;
	    end;
	    Dir := True;
	 elsif "DB" = "" & Incard(Mnemo(0)..Mnemo(1)) then
	    for K in 1..Nextarg-1 loop
	       if Args(K).T/=Int and Args(K).T/=Ch then
		  Asm_Error("DB needs decimal or hex numbers or character to put into memory.");
	       else
		  M(Asm) := Byte(Args(K).Hval mod 256);
		  Asm := Asm + 1;
	       end if;
	    end loop;
	    Dir := True;
	 elsif "CSECT" = "" & Incard(Mnemo(0)..Mnemo(1)) then
	    Dir := True;
	    null; -- nothing to do, but should enable start of code generation
	 elsif "TERM" = "" & Incard(Mnemo(0)..Mnemo(1)) then	
	    if Disp<0 or else Disp>2**16-1 then Disp := 0; end if;
	    if Rr<0 then Rr := 0; end if;
	    if Rb<0 then Rb := 0; end if;
	    Assemble_Ins( Opid => "HALT", -- this must be changed to CSV!
			  Barg0 => 0,
			  Rarg0 => 0,
			  Disp0 => Natural(Disp),
			  I => Ifl,
			  X => Xfl);
	    Dir := True;
	 elsif "ABORT" = "" & Incard(Mnemo(0)..Mnemo(1)) then	
	    Assemble_Ins( Opid => "HALT", -- this must be changed to CSV!
			  Barg0 => 0,
			  Rarg0 => 0,
			  Disp0 => Natural(Half(Disp) or 16#0080#),
			  I => Ifl,
			  X => Xfl);
	    Dir := True;
	 end if;
      end if;
      
      -- INSTRUCTIONS
      if not Dir and Mnemo(0) > 0 then
	 if Nextarg>1 and then Args(1).T = Lbl then
	    if Find_In_Table(Incard(Args(1).Val(0)..Args(1).Val(1))) = Symbol_Not_Found then
	       Add_To_Lt(Incard(Args(1).Val(0)..Args(1).Val(1)),Asm+2);
	       Add_To_Table(Incard(Args(1).Val(0)..Args(1).Val(1)),Word(Asm+2));
	       --	       Put_Line("Added to LT " & Incard(Args(1).Val(0)..Args(1).Val(1)) & " @ " & To_Hex_Half(Asm+2));
	    end if;
	 end if;
	 --	 if Nextarg>2 and then Args(2).T = Lbl then -- but when it could be the second argument!? Never!
	 --	    Add_To_Lt(Incard(Args(2).Val(0)..Args(2).Val(1)),Asm+2);
	 --	    Add_To_Table(Incard(Args(1).Val(0)..Args(1).Val(1)),Asm+2);
	 --	    Put_Line("Added to LT " & Incard(Args(1).Val(0)..Args(1).Val(1)) & " @ " & To_Hex_Half(Asm+2));
	 --	 end if;
	 -- not clear this is ok...
	 Parse_Instruction_Args;
	 if Argerr then
	    Asm_Error("argerr");
	 else 
	    if Rr=-1 then
	       Rr := 0;
	    end if;
	    if Rb=-1 then
	       Rb := 0;
	    end if; -- (o fi bine?)
	 end if;
	 if Xfl then
	    if R(0) > 0 and then Incard(R(0))/='0' then
	       Asm_Error("Wrong instruction, nonzero R specified in indexed mode.");
	    else
	       Assemble_Ins( Opid => Incard(Mnemo(0)..Mnemo(1)),
			     Barg0 => Bfld(Rb),
			     Rarg0 => Rr,
			     Disp0 => Natural(Disp),
			     I => Ifl,
			     X => Xfl);
	    end if;
	 else
	    if R(0) > 0 then
	       Assemble_Ins( Opid => Incard(Mnemo(0)..Mnemo(1)),
			     Barg0 => Bfld(Rb),
			     Rarg0 => Integer'Value(Incard(R(0)..R(1))),
			     Disp0 => Natural(Disp),
			     I => Ifl,
			     X => Xfl);
	    else
	       Assemble_Ins( Opid => Incard(Mnemo(0)..Mnemo(1)),
			     Barg0 => Bfld(Rb),
			     Rarg0 => 0,
			     Disp0 => Natural(Disp),
			     I => Ifl,
			     X => Xfl);
	    end if;
	 end if;
      else
	 null; -- it is just a label, do not assemble anything
      end if;
   end Assemble_Card;
   
   -- THE SIMULATOR
   
   procedure Clear_Memory is
      
   begin
      M := (others => 0);
   end Clear_Memory;
   
   procedure Vidage is
      
      procedure Put_Mem_Line(A: Half) is
	 
      begin
	 for K in 0..7 loop
 	    Put(To_Hex_Byte(M(4*Half(K)+A)));
	    Put(To_Hex_Byte(M(4*Half(K)+A+1)));
	    Put(To_Hex_Byte(M(4*Half(K)+A+2)));
	    Put(To_Hex_Byte(M(4*Half(K)+A+3)));
	    Put(' ');
	 end loop;
	 New_Line;
      end Put_Mem_Line;
      
      Adr: Half;
      
   begin
      Put("ETAT PROGRAMME         ");
      for K in 0..7 loop
	 Put(To_Hex_Byte(Pswbytes(Half(K))));
	 if K=3 then
	    Put(' ');
	 end if;
      end loop;
      
      New_Line(2);
      Put("REGISTRES      ");
      for K in 0..7 loop
	 Put('R' & Integer'Image(K) & "      ");
      end loop;
      New_Line;
      Put("            ");
      Put_Mem_Line(0);
      New_Line;	
      Put("               ");
      for K in 8..15 loop
	 Put('R' & Integer'Image(K) & "     ");
	 if K<10 then
	    Put(' ');
	 end if;
      end loop; 
      New_Line;
      Put("            ");
      Put_Mem_Line(32);
      New_Line(2);	
      Put_Line("VIDAGE MEMOIRE");
      New_Line;
      Adr := 64;
      
      for J in 1..20 loop
	 Put("  ");
	 Put('0');
	 Put(To_Hex_Half(Adr));
	 Put("     ");
	 Put_Mem_Line(Adr);
	 Adr := Adr + 32;
	 if J mod 5 = 0 then
	    New_Line;
	 end if;
      end loop;
      
   end Vidage;
   
   procedure Link_Edit is
      
      Symfound: Natural;
      
   begin
      if Lt_Next < 2 then
	 return; -- nothing to do
      end if;
      for L in 1..Lt_Next-1 loop
	 Symfound := Find_In_Table(Lt(L).Label);
	 if Symfound = 0 then
	    Asm_Error("Symbol " & Lt(L).Label & " not found at link time.");
	 end if;
	 M(Lt(L).Location) := Byte(St(Symfound).Value mod 256);
	 M(Lt(L).Location + 1) := Byte((St(Symfound).Value mod 2**16) / 2**8);
	 New_Line;
	 --	 Put(To_Hex_Half(Lt(L).Location) & " => " & Lt(L).Label  & "(" & To_Hex_Half(Half(St(Symfound).Value mod 2**16)));
	 -- link edited
      end loop;
      
   end Link_Edit;
   
   
   procedure Print_Links_Table is
      
      Symfound: Natural;
      
   begin
      if Lt_Next < 2 then
	 return; -- nothing to do
      end if;
      Put(Heading(1, "LINKS TABLE", 1, 1));
      Put_Line("AT      LABEL     LINK");
      for L in 1..Lt_Next-1 loop
	 Symfound := Find_In_Table(Lt(L).Label);
	 if Symfound = 0 then
	    Put_Line(To_Hex_Half(Lt(L).Location) & " => " & Lt(L).Label  & " (" & "NOT FOUND" & ")");
	 else
	    Put_Line(To_Hex_Half(Lt(L).Location) & " => " & Lt(L).Label  & " (" & To_Hex_Half(Half(St(Symfound).Value mod 2**16)) & ")");
	 end if;
      end loop;
      New_Line(2);
   end Print_Links_Table;
   
   --   Sumdel: Long_Integer; 
   Steps: Long_Integer:= 0;
   Start, Stop: Time;
   Expected_Duration: Duration;
   
   procedure Step is
      
      A: Half; -- the computed address
      Opreg: Register;
      Nrp: Natural:= 0;
      
      procedure Error(M: String) is
	 
      begin
	 Put_Line("Execution error at address " & To_Hex_Half(pc-4) & " : " & M);
	 In_Error := True;
	 return ;
      end Error;
      
      Format: String(1..2);
      Zsdc: Qfld:= Qfld(Natural(Psw.Z)*8 + Natural(Psw.S)*4 + Natural(Psw.D)*2 + Natural(Psw.C));
      
      Ij: Integer;
      Lj: Long_Integer;
      Hj: Half;
      
      Ndep: Byte;
      Modep: Nibble;
      
      procedure Setsz is
	 
      begin
	 if Ri(Half(Ins.Q)) < 0 then Psw.S := 1; else Psw.S := 0; end if;
	 if Ri(Half(Ins.Q)) = 0 then Psw.Z := 1; else Psw.Z := 0; end if;	 
      end;
      
      procedure Setszf is
	 
      begin
	 if Rf(Half(Ins.Q)) < 0.0 then Psw.S := 1; else Psw.S := 0; end if;
	 if Rf(Half(Ins.Q)) = 0.0 then Psw.Z := 1; else Psw.Z := 0; end if;	 
      end;
      
      procedure Setszf8 is
	 
      begin
	 if Rg(Half(Ins.Q)) < 0.0 then Psw.S := 1; else Psw.S := 0; end if;
	 if Rg(Half(Ins.Q)) = 0.0 then Psw.Z := 1; else Psw.Z := 0; end if;	 
      end;
      
      procedure Setsz8 is
	 
      begin
	 if Rl(Half(Ins.Q)) < 0 then Psw.S := 1; else Psw.S := 0; end if;
	 if Rl(Half(Ins.Q)) = 0 then Psw.Z := 1; else Psw.Z := 0; end if;	 
      end;
      
      procedure Setsz2 is
	 
      begin
	 if Rh(Half(Ins.Q))(0) >= Half(2**15) then
	    Psw.S := 1;
	 else
	    Psw.S := 0;
	 end if;
	 if Rh(Opreg)(0)=0 then
	    Psw.Z := 1;
	 else
	    Psw.Z := 0;
	 end if;
      end;
      
      procedure Supervisor is
	 
	 procedure Cond_Assert(Cond: Boolean; Msg: String) is
	    
	 begin
	    if Cond then
	       if Show_All_Assertions then
		  Put_Line("[  OK  ] " & To_Hex_Half(Pc-4) & " assertion holds.");
	       end if;
	    else
	       if  Show_All_Assertions or Show_Failed_Assertions then 
		  Put_Line("[ ???? ] " & To_Hex_Half(Pc-4) & " assertion failed: " & Msg);
	       end if;
	       N_Failed_Asserts := N_Failed_Asserts + 1;
	    end if;
	 end Cond_Assert;
	 
	 Zsdc: Qfld:= Qfld(Natural(Psw.Z)*8 + Natural(Psw.S)*4 + Natural(Psw.D)*2 + Natural(Psw.C));

      begin
	 if Ins.X=1 then -- this is a host instruction
	    case Ins.D is
	       when Assert_Hostop =>
		  for K in 1..Next_Asr-1 loop
		     if Asr(K).Adr=Pc-4 then
			N_Asserts := N_Asserts + 1;
			case Asr(K).Assertt is
			   when Int_Word_eq => 
			      if Ri(Half(Asr(K).Reg)) = Asr(K).Ival then
				 if Show_All_Assertions then
				    Put_Line("[  OK  ] " & To_Hex_Half(Pc-4) & " assertion holds.");
				 end if;
			      else
				 if Show_All_Assertions or Show_Failed_Assertions then 
				    Put_Line("[ ???? ] " & To_Hex_Half(Pc-4) & " assertion:  R" & Dprintf(Integer(Asr(K).Reg),2) & 
				      " = " & Integer'Image(Ri(Half(Asr(K).Reg))) & " /= " & Integer'Image(Asr(K).Ival) & " (I:"
                                      & Long_Integer'Image(Steps) & ")");
				 end if;
				 N_Failed_Asserts := N_Failed_Asserts + 1;
			      end if;
			   when Float_Eq =>
			      if Rf(Half(Asr(K).Reg)) = Asr(K).Fval then -- we should have an epsilon!
				 if Show_All_Assertions then
				    Put_Line("[  OK  ] " & To_Hex_Half(Pc-4) & " assertion holds.");
				 end if;
			      else
				 if Show_All_Assertions or  Show_Failed_Assertions then 
				    Put_Line("[ ???? ] " & To_Hex_Half(Pc-4) & " assertion:  R" & Qfld'Image(Asr(K).Reg) & "(" & 
					       Float'Image(Rf(Half(Asr(K).Reg))) & ") /= " & Float'Image(Asr(K).Fval));
				 end if;
				 N_Failed_Asserts := N_Failed_Asserts + 1;
			      end if;
			   when Double_Eq =>
			      if Rg(Half(Asr(K).Reg)/2) = Asr(K).Dval then -- epsilon please?
				 if Show_All_Assertions then
				    Put_Line("[  OK  ] " & To_Hex_Half(Pc-4) & " assertion holds.");
				 end if;
			      else
				 if  Show_All_Assertions or Show_Failed_Assertions then 
				    Put_Line("[ ???? ] " & To_Hex_Half(Pc-4) & " assertion:  R" & Qfld'Image(Asr(K).Reg) & "(" & 
					       Long_Float'Image(Rg(Half(Asr(K).Reg)/2)) & ") /= " & Long_Float'Image(Asr(K).Dval));
				 end if;
				 N_Failed_Asserts := N_Failed_Asserts + 1;
			      end if;
			   when C =>
			      Cond_Assert(Psw.C=1, " C condition flag is not set.");
			   when NC =>
			      Cond_Assert(Psw.C=0, " C condition flag is set.");
			   when S =>
			      Cond_Assert(Psw.C=1, " S condition flag is not set.");
			   when NS =>
			      Cond_Assert(Psw.C=0, " S condition flag is set.");
			   when Z =>
			      Cond_Assert(Psw.C=1, " Z condition flag is not set.");
			   when NZ =>
			      Cond_Assert(Psw.C=0, " Z condition flag is set.");
			   when D =>
			      Cond_Assert(Psw.C=1, " D condition flag is not set.");
			   when ND =>
			      Cond_Assert(Psw.C=0, " D condition flag is set.");
			   when CF =>
			      Cond_Assert((Qfld(Asr(K).Ival) and Zsdc)=0, " PSW conditions, " & Qfld'Image(Zsdc) &
					    " are not inverse to asserted " & Integer'Image(Asr(K).Ival));
			   when CT =>
			      Cond_Assert((Qfld(Asr(K).Ival) and Zsdc)/=0, " PSW conditions, " & Qfld'Image(Zsdc) &
					    " are not all equal to asserted " & Integer'Image(Asr(K).Ival));
			   when others => 
			      if Show_All_Assertions or Show_Failed_Assertions then 
				 Put_Line("[ ???? ] " & To_Hex_Half(Pc-4) & " assertion:  R" & 
					    Qfld'Image(Asr(K).Reg) & "Invalid assertion.");
			      end if;
			      N_Failed_Asserts := N_Failed_Asserts + 1;
			end case;
		     end if;
		  end loop;
	       when others =>
		  Error("Invalid hostop.");
	    end case;
	 else
	    Error("Non hostop supervisor instructions not implemented yet.");
	 end if;
      end Supervisor;
      
      
      
   begin
      --      Put(To_Hex_Half(Pc) & " ");
      Inb := M(Pc..Pc+3);
      Pc := Pc+4;
      Opreg := Register(Ins.Q); 
      -- the operand register must be 0
      -- in indexed addressing mode
      
      if Ins.X=0 and Ins.I=0 then
	 if Ins.B=0 then
	    A := Ins.D;
	 else
	    A := Ins.D + Rh(Register(Ins.B)+8)(0);
	 end if;
      elsif Ins.X=0 and Ins.I=1 then
	 if Ins.B=0 then
	    A := Ins.D;
	 else
	    A := Ins.D + Rh(Register(Ins.B)+8)(0);
	 end if;
	 A := Half(M(A)) + 256*Half(M(A+1));
      elsif Ins.X=1 and Ins.I=0 then
	 if Ins.B=0 then
	    A := Ins.D + Rh(Register(Ins.Q))(0);
	 else
	    A := Ins.D + Rh(Register(Ins.B)+8)(0) + Rh(Register(Ins.Q))(0);
	 end if;
	 Opreg := 0;
      elsif Ins.X=1 and Ins.I=1 then
	 if Ins.B=0 then
	    A := Ins.D;
	 else
	    A := Ins.D + Rh(Register(Ins.B)+8)(0);
	 end if;
	 A := Half(M(A)) + 256*Half(M(A+1));
	 A := A + Rh(Register(Ins.Q))(0);
	 Opreg := 0;
      else
	 Error("Addressing mode not supported implemented so far.");
	 return;
      end if;
      
      -- get the bit fields for shifts anyway
      
      Ndep := Byte(A mod 2**6);
      Modep := Nibble((A mod 2**12) / 2**8);
      
      -- first, we wait...
      
      Expected_Duration := Expected_Duration + Timings(Byopcod(Ins.Op));
      
      case Timing_Conf is
	 when Original =>
	    while Start+Expected_Duration-Clock > 0.001 loop
	       delay 0.001;
	    end loop;
	    --	    Sumdel := Long_Integer(Long_Increments_Per_Microsecond  * Long_Float(Timings(Byopcod(Ins.Op))) / 1000.0);
	    --	    Put_Line(Long_Float'Image(Long_Increments_Per_Microsecond) & Long_Integer'Image(Sumdel));
	    --	    while Sumdel > 0 loop
	    --	       Sumdel := Sumdel - 1;
	    --	    end loop;
	    null;
	 when Native =>
	    null;
      end case;
      
      case Byopcod(Ins.Op) is
	 when Ad4 =>
	    Ri(Opreg) := Ri(Opreg) + Memread4(A);
	    Setsz;
	    --	      Words_Register(Word(M(A)) + 2**8 * Word(M(A+1)) + 2**16 * Word(M(A+2)) + 2**24 * Word(M(A+3)));
	 when Ad4i =>
	    Rw(Opreg) := Words_Register(Word(Rw(Opreg)) + Word(A));
	    Setsz;
	 when Ad8 =>
	    Rl(Opreg/2) := Rl(Opreg/2) + Memread8(A);
	    Setsz8;
	 when Adf8 =>
	    Rg(Opreg/2) := Rg(Opreg/2) + Memread8(A);
	    Setszf8;
	 when Adf4 =>
	    Rf(Opreg) := Rf(Opreg) + Memread4(A);
	    Setszf;
	 when Adh2 =>
	    Rs(Opreg)(0) := Rs(Opreg)(0) + Memread2(A);
	    Setsz2;
	 when Bru =>
	    Pc := A;
	 when Bcf =>
	    if (Ins.Q and Zsdc) = 0 then -- perhaps something wrong?
	       Pc := A;
	    end if;
	 when Bct =>
	    if (Ins.Q and Zsdc) /= 0 then
	       Pc := A;
	    end if;
	 when Bal =>
	    Rw(Opreg) := Words_Register(Pc);
	    Pc := A;
	 when Cp1i =>
	    if Rb(Opreg)(0) = Byte(A mod 256) then
	       Psw.Z := 1;
	    else
	       Psw.Z := 0;
	    end if; -- ar trebui si altele dar sunt signed?
	    if Rb(Opreg)(0) < Byte(A mod 256) then
	       Psw.S := 1;
	    else
	       Psw.S := 0;
	    end if; -- ar trebui si altele dar sunt signed?
	 when Cp1 =>
	    if Rb(Opreg)(0) = M(A) then
	       Psw.Z := 1;
	    else
	       Psw.Z := 0;
	    end if; -- ar trebui si altele dar sunt signed?
	    if Rb(Opreg)(0) < M(A) Then
	       Psw.S := 1;
	    else
	       Psw.S := 0;
	    end if; -- ar trebui si altele dar sunt signed?
	 when Cp2 =>
	    if Rh(Opreg)(0) = Half(M(A)) + 256 * Half(M(A+1)) then
	       Psw.Z := 1;
	    else
	       Psw.Z := 0;
	    end if; -- ar trebui si altele dar sunt signed?
	    if Rh(Opreg)(0) < Half(M(A)) + 256 * Half(M(A+1)) then
	       Psw.S := 1;
	    else
	       Psw.S := 0;
	    end if; -- ar trebui si altele dar sunt signed?
	 when Cp4 =>
	    Ij := Integer(Rw(Opreg)) - 
	      Integer(Words_Register(Word(M(A)) + 2**8 * Word(M(A+1)) + 2**16 * Word(M(A+2)) + 2**24 * Word(M(A+3))));
	    if Ij=0 then
	       Psw.Z := 1;
	    else
	       Psw.Z := 0;
	    end if;
	    if Ij<0 then
	       Psw.S := 1;
	    else
	       Psw.S := 0;
	    end if;
	 when Cp8 =>
	    Lj := Long_Integer(Rl(Opreg) - Memread8(A));
	    if Lj=0 then
	       Psw.Z := 1;
	    else
	       Psw.Z := 0;
	    end if;
	    if Lj<0 then
	       Psw.S := 1;
	    else
	       Psw.S := 0;
	    end if;
	 when Csv =>
	    Supervisor;
	 when Dc4 =>
	    Ij := Memread4(A);
	    Ij := Ij - Integer(Ins.Q);
	    Memstor4(A,Ij);
	    if Ij=0 then
	       Psw.Z := 1;
	    else
	       Psw.Z := 0;
	    end if;
	    if Ij<0 then
	       Psw.S := 1;
	    else
	       Psw.S := 0;
	    end if;
	 when Dv2 =>
	    Ri(Opreg) := Ri(Opreg) / Integer(Memread2(A));
	    Setsz2;
	 when Dvu2 =>
	    Rw(Opreg) := Rw(Opreg) / Words_Register(Half(M(A)) + 256 * Half(M(A+1)));
	    Setsz2;
	 when Dv4 =>
	    Ri(Opreg) := Ri(Opreg) / Memread4(A);
	    Setsz;
	    --	 when Dv8 => -- does not exist
	    --	    Rl(Opreg/2) := Rl(Opreg/2) / Memread8(A);
	    --sz?
	 when Dvf8 =>
	    Rg(Opreg/2) := Rg(Opreg/2) / Memread8(A);
	    Setszf8;
	 when Dvf4 =>
	    Rf(Opreg) := Rf(Opreg) / Memread4(A);
	    Setszf;
	 when Eo2 =>
	    Rh(Opreg)(0) := Rh(Opreg)(0) xor (Half(M(A)) + 256 * Half(M(A+1)));
	    Setsz2;
	 when Eo4 =>
	    Rw(Opreg) :=  Rw(Opreg) xor 
	      Words_Register(Word(M(A)) + 2**8 * Word(M(A+1)) + 2**16 * Word(M(A+2)) + 2**24 * Word(M(A+3)));
	    Setsz;
	 when Ex2 =>
	    Rh(Opreg)(0) := Rh(Opreg)(0) and (Half(M(A)) + 256 * Half(M(A+1)));
	    Setsz2;
	 when Ex4 =>
	    Rw(Opreg) :=  Rw(Opreg) and
	      Words_Register(Word(M(A)) + 2**8 * Word(M(A+1)) + 2**16 * Word(M(A+2)) + 2**24 * Word(M(A+3)));
	    Setsz;
	 when Ic2 =>
	    Hj := Half(M(A) + (2**8)*M(A+1));
	    Hj := Hj + Half(Ins.Q);
	    M(A+1) := Byte(Hj / (2**8));
	    M(A) := Byte(Hj mod (2**8));
	    if Rs(A)(0)=0 then
	       Psw.Z := 1;
	    else
	       Psw.Z := 0;
	    end if;
	    if Rs(A)(0)<0 then
	       Psw.S := 1;
	    else
	       Psw.S := 0;
	    end if;
	 when Ic4 =>
	    Ij := Memread4(A);
	    Ij := Ij + Integer(Ins.Q);
	    Memstor4(A,Ij);
	    if Ij=0 then
	       Psw.Z := 1;
	    else
	       Psw.Z := 0;
	    end if;
	    if Ij<0 then
	       Psw.S := 1;
	    else
	       Psw.S := 0;
	    end if;
	 when Ldc2 => -- cred ca e gresit, trebuie - Rh(0) sau ceva
	    Rs(Opreg)(0) := - Rs(A)(0); -- (was) not (Half(M(A)) + 256 * Half(M(A+1)));
	    Setsz2;
	 when Ldc4 => -- gresit, trebuie - Ri
	    Ri(Opreg) :=  -Ri(A); -- not Words_Register(Word(M(A)) + 2**8 * Word(M(A+1)) + 2**16 * Word(M(A+2))+2**24 * Word(M(A+3)));
	    Setsz;
	 when Ld1i =>
	    Rb(Opreg)(0) := Byte(Half(A) mod 256);
	    if Rb(Opreg)(0)=0 then
	       Psw.Z := 1;
	    else
	       Psw.Z := 0;
	    end if;
	 when Ld2i =>
	    Rh(Opreg)(0) := Half(A);
	    Setsz2;
	 when Ld1 =>
	    Rb(Opreg)(0) := M(A);
	    if Rb(Opreg)(0)=0 then
	       Psw.Z := 1;
	    else
	       Psw.Z := 0;
	    end if;
	 when Ldh2 =>
	    Rb(Opreg)(0) := M(A);
	    Rb(Opreg)(1) := M(A+1);
	    Setsz2;
	 when Ldl2 =>
	    Rb(Opreg)(2) := M(A);
	    Rb(Opreg)(3) := M(A+1);
	 when Ld4 =>
	    Rb(Opreg) := Bytes_Register(M(A..A+3));
	    Setsz;
	 when Ld8 =>	
	    Rb(Opreg) := Bytes_Register(M(A..A+3));
	    Rb(Opreg+1) := Bytes_Register(M(A+4..A+7));
	    Setsz8;
	 when Ldm =>
	    for Y in 0..Ins.Q loop
	       Rb(Half(Y)) := Bytes_Register(M((A+Half(4*Y))..(A+Half(4*Y+3))));
	    end loop;
	 when Ld4i =>
	    Ri(Opreg) := Integer(A);
	    Setsz;
	 when Mg2 =>
	    Rh(Opreg)(0) := Rh(Opreg)(0) or (Half(M(A)) + 256 * Half(M(A+1)));
	    Setsz2;
	 when Mg4 =>
	    Rw(Opreg) :=  Rw(Opreg) or 
	      Words_Register(Word(M(A)) + 2**8 * Word(M(A+1)) + 2**16 * Word(M(A+2)) + 2**24 * Word(M(A+3)));
	    Setsz;
	 when Mp2 =>
	    Ri(Opreg) := Ri(Opreg) * Integer(Memread2(A));
	    Setsz;
	 when Mpu2 =>
	    Rw(Opreg) := Rw(Opreg) * Words_Register(Half(M(A)) + 256 * Half(M(A+1)));
	    Setsz;
	 when Mp4 =>
	    Ri(Opreg) := Ri(Opreg) * Memread4(A);
	    Setsz;
	 when Mpf8 =>
	    Rg(Opreg/2) := Rg(Opreg/2) * Memread8(A);
	    Setszf8;
	 when Mpf4 =>
	    Rf(Opreg) := Rf(Opreg) * Memread4(A);
	    Setszf;	
	 when Nf4 =>
	    Error("NF4 not implemented yet.");
	    null;
	 when Sb4i =>
	    Rw(Opreg) := Rw(Opreg) - Words_Register(A);
	    Setsz;
	 when Sb4 =>
	    Ri(Opreg) := Ri(Opreg) - Memread4(A);
	    Setsz;
	 when Sb8 =>
	    Rl(Opreg/2) := Rl(Opreg/2) - Memread8(A);
	    Setsz8;
	 when Sbh2 =>
	    Rs(Opreg)(0) := Rs(Opreg)(0) - Memread2(A);
	    Setsz2;
	 when Sh2 =>
	    case Modep is
	       when 0 => -- SLL logical shift left
		  Rh(Opreg)(0) := Rh(Opreg)(0) * 2**Natural(Ndep);
		  Setsz2; -- assumes opreg is ins.q
	       when 4 => -- SRL Shift Right Logical
		  Rh(Opreg)(0) := Rh(Opreg)(0) / 2**Natural(Ndep);
		  Rh(Opreg)(0) := Rh(Opreg)(0) and (2**Natural(16-Ndep) - 1);
		  Setsz2;
	       when 1 => -- SLC closed shift left (rotate left)
		  Rh(Opreg)(0) := (Rh(Opreg)(0) * 2**Natural(Ndep)) or (Rh(Opreg)(0) / 2**Natural(16-Ndep));
		  Setsz2;
	       when 5 => -- SRC Shift Right Closed (rotate right)
		  Rh(Opreg)(0) := Rh(Opreg)(0) / 2**Natural(Ndep);
		  Rh(Opreg)(0) := (Rh(Opreg)(0) and (2**Natural(16-Ndep) - 1)) or 
		    ((Rh(Opreg)(0) mod (2**Natural(Ndep))) * 2**Natural(16-Ndep));
		  Setsz2;
	       when others => 
		  Error("Not implemented yet.");
	    end case;
	 when Sh4 =>
	    case Modep is
	       when 0 => -- SLL logical shift left
		  Rw(Opreg) := Rw(Opreg) * 2**Natural(Ndep);
		  Setsz;
	       when 4 => -- SRL Shift Right Logical
		  Rw(Opreg) := Rw(Opreg) / 2**Natural(Ndep);
		  Rw(Opreg) := Rw(Opreg) and (2**Natural(32-Ndep) - 1);
		  Setsz;
	       when 1 => -- SLC closed shift left (rotate left)
		  Rw(Opreg) := (Rw(Opreg) * 2**Natural(Ndep)) or (Rw(Opreg) / 2**Natural(32-Ndep));
		  Setsz;
	       when 5 => -- SRC Shift Right Closed (rotate right)
		  Rw(Opreg) := Rw(Opreg) / 2**Natural(Ndep);
		  Rw(Opreg) := (Rw(Opreg) and (2**Natural(32-Ndep) - 1)) or 
		    ((Rw(Opreg) mod (2**Natural(Ndep))) * 2**Natural(32-Ndep));
		  Setsz;
	       when others => 
		  Error("Not implemented yet.");
	    end case;
	 when Sh8 =>
	    case Modep is
	       when 0 => -- SLL logical shift left
		  Ru(Opreg/2) := Ru(Opreg/2) * 2**Natural(Ndep);
		  Setsz8;
	       when 4 => -- SRL Shift Right Logical
		  Ru(Opreg/2) := Ru(Opreg/2) / 2**Natural(Ndep);
		  Ru(Opreg/2) := Ru(Opreg/2) and (2**Natural(64-Ndep) - 1);
		  Setsz8;
	       when 1 => -- SLC closed shift left (rotate left)
		  Ru(Opreg/2) := (Ru(Opreg/2) * 2**Natural(Ndep)) or (Ru(Opreg/2) / 2**Natural(64-Ndep));
		  Setsz8;
	       when 5 => -- SRC Shift Right Closed (rotate right)
		  Ru(Opreg/2) := Ru(Opreg/2) / 2**Natural(Ndep);
		  Ru(Opreg/2) := (Ru(Opreg/2) and (2**Natural(64-Ndep) - 1)) or 
		    ((Ru(Opreg/2) mod (2**Natural(Ndep))) * 2**Natural(64-Ndep));
		  Setsz8;
	       when others => 
		  Error("Not implemented yet.");
	    end case;
	 when Sbf8 =>
	    Rg(Opreg/2) := Rg(Opreg/2) - Memread8(A);
	    Setszf8;
	 when Sbf4 =>
	    Rf(Opreg) := Rf(Opreg) - Memread4(A);
	    Setszf;
	 when St1 => 
	    M(A) := Rb(Opreg)(0);
	    if Rb(Opreg)(0)=0 then
	       Psw.Z := 1;
	    else
	       Psw.Z := 0;
	    end if;
	 when St4 =>
	    M(A..A+3) := Memory_Bytes(Rb(Opreg));
	    Setsz;
	 when St8 =>
	    M(A..A+3) := Memory_Bytes(Rb(Opreg));
	    M(A+4..A+7) := Memory_Bytes(Rb(Opreg+1));
	    Setsz8;
	 when Sth2 =>
	    M(A) := Rb(Opreg)(0); -- i would have expected indices 2,3 but bocsa seems to disagree
	    M(A+1) := Rb(Opreg)(1);
	    Setsz2;
	 when Stm =>
	    for Y in 0..Ins.Q loop
	       M((A+Half(4*Y))..(A+Half(4*Y+3))) := Memory_Bytes(Rb(Half(Y)));
	    end loop;
	 when Halt =>
	    Halted := True;
	 when Print => -- this should be a case..
	    Format(1) := Character'Val(Natural(Ins.D/256));
	    Format(2) := Character'Val(Natural(Ins.D mod 256));
	    if Ins.D=0  then
	       Put(Character'Val(Rw(Opreg) and 16#0ff#));
	    elsif Format(1)='\' and then Format(2)='n' then
	       New_Line;
	    elsif Ins.D<255 and then Format(2)='d' then
	       Put(Integer'Image(Ri(Opreg)));
	    elsif Ins.D<255 and then Format(2)='f' then
	       Put(Float'Image(Rf(Opreg)));
	    elsif Ins.D<255 and then Format(2)='g' then
	       Put(Long_Float'Image(Rg(Opreg/2)));
	    elsif Ins.D<255 and then Format(2)='x' then	 
	       Put(To_Hex_Byte(Rb(Opreg)(3)));
	       Put(To_Hex_Byte(Rb(Opreg)(2)));
	       Put(To_Hex_Byte(Rb(Opreg)(1)));
	       Put(To_Hex_Byte(Rb(Opreg)(0)));
	    else
	       if Format(1)<' ' or Format(1)>'~'then
		  Format(1) := '?';
	       end if;
	       if Format(2)<' ' or Format(2)>'~' then
		  Format(2) := '?';
	       end if;
	       Put(Format);
	    end if;
	 when others =>
	    Error("BAD OP, not implemented (yet)." & Opcod'Image(Ins.Op));
      end case;
      return;
   end Step;

   procedure Run(Maxtime: Long_Integer) is

   begin
      -- here, could initialise nr of assertions etc
      if Show_Run_Output_Limits then
	 Put_Line("RUNNING FROM " & 
		    To_Hex_Half(Run_Entry) & " ===>");
      end if;
      Halted := False;
      Pc := Run_Entry;
      Start := Clock;
      Jrunstart := Start;
      for K in 1..Maxtime loop
	 Step;
	 Steps := K;
	 if Halted or In_Error Then
	    exit;
	 end if;
      end loop;
      Stop := Clock;
      Jrunend := Stop;
      if Show_Run_Output_Limits then
	 Put_Line("<=== FINISHED ");
      end if;
      
      if Show_Exit_Reason then
	 if Halted then
	    Put_Line("HALT instruction encountered.");
	 elsif In_Error then
	    Put_Line("Stopped because of error.");
	 else
	    Put_Line("Finished by the system after the " & Long_Integer'Image(Maxtime) & " steps.");
	 end if;
      end if;
      
      if Show_Run_Statistics then
	 Put("Executed " & Long_Integer'Image(Steps) & " instructions " & " in " & Duration'Image(Stop-Start)
	       & " seconds, " & " at a speed of ");
	 Put(Long_Float(Steps) / Long_Float(Stop-Start) / 1_000_000.0, 3, 6, 0);
	 Put_Line(" BogoMIPS");
	 Put_Line(Natural'Image(Nr_Exec_Errors) & 
		    " runtime errors encountered, maximum level: " & 
		    Natural'Image(Max_Exec_Error_Level));
	 Put_Line(Natural'Image(N_Asserts) & 
		    " assertions tested, of which: " & 
		    Natural'Image(N_Failed_Asserts) & 
		    " failed.");
      end if;
      --      Vidage;
   end Run;
   
   -- PROCESS JOB SET
   
   procedure Process_Control_Card is
      
      Cmd: String := "" & Incard(Mnemo(0)..Mnemo(1));
      Cc_Done: Boolean:= False;
      Maxtime: Long_Integer := 500;
      
      procedure Process_Conf_Args is
	 
      begin
	 for K in 1..Nextarg-1 loop
	    if Args(K).Name(0) > 0 and then
	      Incard(Args(K).Name(0)..Args(K).Name(1)) = "VERB"
	      and then Args(K).T=Int  then
	       Set_Verbosity(Args(K).Ival);
	    end if;
	    if Args(K).Name(0) > 0 and then
	      Incard(Args(K).Name(0)..Args(K).Name(1)) = "PN"
	      and then Args(K).T=Lbl  then
--	       Put_Line("read pn=" & Incard(Args(K).Val(0)..Args(K).Val(1)));
	       Pn(1..Args(K).Val(1)-Args(K).Val(0)+1) := Incard(Args(K).Val(0)..Args(K).Val(1));
	    end if;
	    if Args(K).Name(0) > 0 and then
	      Incard(Args(K).Name(0)..Args(K).Name(1)) = "AN"
	      and then Args(K).T=Lbl  then
--	       Put_Line("read pn=" & Incard(Args(K).Val(0)..Args(K).Val(1)));
	       An(1..Args(K).Val(1)-Args(K).Val(0)+1) := Incard(Args(K).Val(0)..Args(K).Val(1));
	    end if;
	    if Args(K).Name(0) > 0 and then
	      Incard(Args(K).Name(0)..Args(K).Name(1)) = "SPEED" then
	       if Args(K).T=Str  or else Args(K).T=Lbl then
		  if Incard(Args(K).Val(0)) = 'N' then -- native, default
		     Timing_Conf := Native;
		  elsif Incard(Args(K).Val(0)) = 'O' then -- original
		     Timing_Conf := Original;
		     Put_Line("speed set to original");
		  else
		     Ctrl_Error("SPEED may be N(ative), or >0, that is fast or O(riginal), or 0, Felix-C-256.");
		  end if;
	       elsif Args(K).T=Int then
		  if Args(K).Ival = 0 then
		     Timing_Conf := Original;
		  else 
		     Timing_Conf := Native;
		  end if;
	       end if;
	    end if;
	 end loop;
	 Cc_Done := True;
      end Process_Conf_Args;
      
   begin
      --      Put_Line("cmd = " & Cmd);
      if Cmd="JOB" then
	 if Injob then
	    Ctrl_Error("Already in job.");
	    return;
	 else
	    Jstart := Clock;
	    -- should check there is at least one job name
	    Jn(1..Args(1).Val(1)-Args(1).Val(0)+1) := Incard(Args(1).Val(0)..Args(1).Val(1));
	    Runn(1..8) := Dprintf(Integer(Float(10**8-1) * Random(Rungen)) mod (10**8 - 1),8);
	    Process_Conf_Args;
	    Print_Job_Banner;
	    Injob := True;
	    Incompile := False;
	    Cc_Done := True;
	 end if;
      end if;
      if Cmd="CONF" then
	 Process_Conf_Args;
	 Cc_Done := True;
      end if;
      if Cmd="SAY" or else Cmd="LIST" then
	 for K in 1..Nextarg-1 loop
	    if Args(K).T=Str and then Args(K).Sval(1..Args(K).Len) = "HELP" then
	       Print_Help;
	    elsif Args(K).T=Str and then Args(K).Sval(1..Args(K).Len) = "ABOUT" then
	       Print_About;
	    elsif Args(K).T=Str and then Args(K).Sval(1..Args(K).Len) = "SYMS" then
	       Print_Symbol_Table;
	    elsif Args(K).T=Str and then Args(K).Sval(1..Args(K).Len) = "LINKS" then
	       Print_Links_Table;
	    elsif Args(K).T=Str and then Args(K).Sval(1..Args(K).Len) = "DUMP" then
	       Vidage;
	    elsif Args(K).Name(0) > 0 and then
	      Incard(Args(K).Name(0)..Args(K).Name(1)) = "MSG"
	      and then Args(K).T=Str  then
	       Put_Line(Args(K).Sval(1..Args(K).Len));
	    end if;
	 end loop;
	 Cc_Done := True;
      end if;
      if Injob then -- only considered inside a job, ignored otherwise
	 if Cmd="EOJ" then
	    Injob := False;
	    Incompile := False;
	    Cc_Done := True;
	    Jend := Clock;
	 elsif Cmd="COMPILE" then
	    Incompile := True;
	    Cc_Done := True;
	    if Show_Assembly then
	       Print_Asm_Header;
	    end if;
	 elsif Cmd="LINK" then
	    --	    Put("LINKING " & Integer'Image(Lt_Next-1) & " LOCATIONS ... ");
	    Incompile := False;
	    Link_Edit;
	    --	    Put_Line("DONE");
	    Cc_Done := True;
	 elsif Cmd="RUN" then
	    for K in 1..Nextarg-1 loop
	       if Args(K).Name(0) > 0 and then
		 Incard(Args(K).Name(0)..Args(K).Name(1)) = "KINS"
		 and then Args(K).T=Int  then
		  Maxtime := Long_Integer(Args(K).Ival) * 1_000;
	       end if;
	    end loop;
	    for K in 1..Nextarg-1 loop
	       if Args(K).Name(0) > 0 and then
		 Incard(Args(K).Name(0)..Args(K).Name(1)) = "INS"
		 and then Args(K).T=Int  then
		  Maxtime := Long_Integer(Args(K).Ival);
	       end if;
	    end loop;
	    Run(Maxtime);
	    Cc_Done := True;
	 end if;
      end if;
      if not Cc_Done then
	 Asm_Error("Control card could not be processed.");
      end if;
   end;

   
   procedure Process_Job_Set(From: File_Type) is
      
   begin
      Set_All_Ops;
      Clear_Memory;
      -- Print_Asm_Header;
      
      -- read and assemble
      
      Asm := 64;
      
      loop
	 Read_Next_Card(From);
	 if No_More_Cards then
	    exit;
	 end if;
	 Parse_Card;
	 if Card_Type='.' then
	    Process_Control_Card;
	 end if;
	 if Incompile then
	    if (Card_Type=' ' or Card_Type='A') and then Incompile then
	       Assemble_Card;
	       if Show_Assembly then
		  Print_Card (not Dir);
	       end if;
	    else
	       if Show_Directives then
		  Print_Card (False);
	       end if;
	    end if;
	 end if;
      end loop;
      
   end Process_Job_Set;
   
   -- SOME EARLY TESTS
   
   procedure Test_Asm_1 is
      
   begin
      Assemble_Ins("LD1I", 1, 0, Character'Pos('H'));
      Assemble_Ins("LD1I", 2, 0, Character'Pos('e'));
      Assemble_Ins("LD1I", 3, 0, Character'Pos('l'));
      Assemble_Ins("LD1I", 4, 0, Character'Pos('l'));
      Assemble_Ins("LD1I", 5, 0, Character'Pos('o'));
      Assemble_Ins("LD1I", 6, 0, Character'Pos(Ascii.Lf));
      Assemble_Ins("PRINT",1,0,0);
      Assemble_Ins("PRINT",2,0,0);
      Assemble_Ins("PRINT",3,0,0);
      Assemble_Ins("PRINT",4,0,0);
      Assemble_Ins("PRINT",5,0,0);
      Assemble_Ins("PRINT",6,0,0);
      Assemble_Ins("HALT",0,0,0);
   end Test_Asm_1;
   
   procedure Test_Ins_Mem is
      
   begin
      Asm := 213;
      Inw := 16#01020304#;
      M(Asm..Asm+3) := Inb; 
      Put_Line(Half'Image(INS.D));
      Inw := 0;
      Put_Line(Half'Image(INS.D));
      Inb := M(Asm..Asm+3);
      Put_Line(Half'Image(INS.D));
   end Test_Ins_Mem;
   
   -- THE MAIN LOOP
   
   
begin
   
   -- calibrate delay loop
   --   pragma Optimize(Off);
   
   Reset(Rungen);
--   Reset(Rungen, Integer((Duration(Clock) - Long_Float'Floor(Long_Float(Clock))) * 1000_000.0) mod 1000_000);
   
   if Argument_Count /= 0 then
      Print_About;
      return;
   end if;
   
   Process_Job_Set(Standard_Input);
   
end Felix;
